# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

from csmpe.plugins import CSMPlugin


class Plugin(CSMPlugin):
    """This plugin retrieves software information from the device."""
    name = "Get Inventory Plugin"
    platforms = {'ASR900'}
    phases = {'Get-Inventory'}
    os = {'IOS'}

    def run(self):
        get_package(self.ctx)
        get_inventory(self.ctx)


def get_inventory(ctx):
    # Save the output of "show inventory"
    output = ctx.send("show inventory")
    ctx.save_data("cli_show_inventory", output)


def get_package(ctx):
    ctx.save_data("cli_show_install_committed",
                  ctx.send("show version | include ^System image"))

    ctx.send('cd flash:')
    ctx.save_data("cli_show_install_inactive", ctx.send("dir"))



import time
import re
import json

from csmpe.context import PluginError
from csmpe.core_plugins.csm_custom_commands_capture.plugin import Plugin as CmdCapturePlugin

SUPPORTED_HW_JSON = "./asr9k_x64/migration_supported_hw.json"

ADMIN_RP = "\d+/RS?P\d+"
ADMIN_LC = "\d+/\d+"


def log_and_post_status(ctx, msg):
    ctx.info(msg)
    ctx.post_status(msg)


def parse_exr_admin_show_platform(output):
    """Get all RSP/RP/LC string node names matched with the card type."""
    inventory = {}
    lines = output.split('\n')

    for line in lines:
        line = line.strip()
        if len(line) > 0 and line[0].isdigit():
            node = line[:10].strip()
            # print "node = *{}*".format(node)
            node_type = line[10:34].strip(),
            # print "node_type = *{}*".format(node_type)
            inventory[node] = node_type
    return inventory


def parse_admin_show_platform(output):
    """
    :param output: output from 'admin show platform' for ASR9K
    :return: list of tuples of (node name, node info)

    ASR9K:
    Node            Type                      State            Config State
    -----------------------------------------------------------------------------
    0/RSP0/CPU0     A9K-RSP440-SE(Active)     IOS XR RUN       PWR,NSHUT,MON
    0/FT0/SP        ASR-9006-FAN              READY
    0/1/CPU0        A9K-40GE-E                IOS XR RUN       PWR,NSHUT,MON
    0/2/CPU0        A9K-MOD80-SE              UNPOWERED        NPWR,NSHUT,MON
    0/3/CPU0        A9K-8T-L                  UNPOWERED        NPWR,NSHUT,MON
    0/PS0/M0/SP     A9K-3KW-AC                READY            PWR,NSHUT,MON
    0/PS0/M1/SP     A9K-3KW-AC                READY            PWR,NSHUT,MON
    """
    inventory = []
    lines = output.split('\n')
    for line in lines:
        line = line.strip()
        if len(line) > 0 and line[0].isdigit():
            node = line[:16].strip()
            entry = {
                'type': line[16:42].strip(),
                'state': line[42:59].strip(),
                'config_state': line[59:].strip()
            }
            inventory.append((node, entry))
            # print node

    return inventory


def get_all_supported_nodes(ctx, supported_cards):
    """Get the list of string node names(all available RSP/RP/LC) that are supported for migration."""
    supported_nodes = []
    ctx.send("admin")
    # show platform can take more than 1 minute after router reload. Issue No. 47
    output = ctx.send("show platform", timeout=600)
    inventory = parse_exr_admin_show_platform(output)

    rp_pattern = re.compile(ADMIN_RP)
    lc_pattern = re.compile(ADMIN_LC)
    supported_rp = supported_cards.get("RP")
    if not supported_rp:
        ctx.error("Missing supported hardware information for RP in {}.".format(SUPPORTED_HW_JSON))

    supported_lc = supported_cards.get("LC")
    if not supported_lc:
        ctx.error("Missing supported hardware information for LC in {}.".format(SUPPORTED_HW_JSON))

    for node, node_type in inventory.items():
        if rp_pattern.match(node):
            for rp in supported_rp:
                if rp in node_type:
                    supported_nodes.append(node)
                    break
        elif lc_pattern.match(node):
            for lc in supported_lc:
                if lc in node_type:
                    supported_nodes.append(node)
                    break
    ctx.send("exit")
    return supported_nodes


def get_version(ctx):
    output = ctx.send("show version | include Version")
    version = re.search("Version\s*?(\d+\.\d+)\.\d+(?:\.\d+I)?", output)
    if not version:
        ctx.error("Failure to retrieve release number.")
    return version.group(1)


def wait_for_final_band(ctx):
    """This is for ASR9K eXR. Wait for all present nodes to come to FINAL Band."""
    exr_version = get_version(ctx)
    with open(SUPPORTED_HW_JSON) as supported_hw_file:
        supported_hw = json.load(supported_hw_file)
    if supported_hw.get(exr_version) is None:
        ctx.error("No hardware support information available for release {}.".format(exr_version))

    supported_nodes = get_all_supported_nodes(ctx, supported_hw.get(exr_version))
    # Wait for all nodes to Final Band
    timeout = 1500
    poll_time = 20
    time_waited = 0

    cmd = "show platform vm"
    # ctx.send("admin")
    # cmd = "show platform"
    while 1:
        # Wait till all nodes are in FINAL Band
        time_waited += poll_time
        if time_waited >= timeout:
            break
        time.sleep(poll_time)
        output = ctx.send(cmd)
        if check_show_plat_vm(output, supported_nodes):
            return True

        """
        if check_sw_status_admin(supported_nodes, output):
            ctx.send("exit")
        """
    # ctx.send("exit")
    # Some nodes did not come to FINAL Band
    return False


def check_show_plat_vm(output, supported_nodes):
    """Check if all supported nodes reached FINAL Band status"""

    all_nodes_ready = True
    entries_in_show_plat_vm = []
    lines = output.splitlines()
    for line in lines:
        line = line.strip()
        if len(line) > 0 and line[0].isdigit():
            entries_in_show_plat_vm.append(line)

    if len(entries_in_show_plat_vm) < len(supported_nodes):
        all_nodes_ready = False
    else:
        for node in supported_nodes:
            node_is_ready = False
            for entry in entries_in_show_plat_vm:
                if node in entry:
                    if 'FINAL Band' in entry:
                        node_is_ready = True
                    break
            if not node_is_ready:
                all_nodes_ready = False
                break
    return all_nodes_ready


def check_sw_status_admin(supported_nodes, output):
    """Check if a node is in OPERATIONAL status"""
    lines = output.splitlines()

    for line in lines:
        line = line.strip()
        if len(line) > 0 and line[0].isdigit():
            node = line[0:10].strip()
            if node in supported_nodes:
                sw_status = line[48:62].strip()
                if "OPERATIONAL" not in sw_status:
                    return False
    return True


def run_additional_custom_commands(ctx, additional_commands):
    """
    Help method to run additional custom commands besides the ones defined by user
    :param ctx: plugin context (PluginContext)
    :param additional_commands: a set containing all the custom commands you wish to run
    :return: None
    """
    user_defined_custom_commands = ctx.custom_commands

    if type(user_defined_custom_commands) is list:
        for command in user_defined_custom_commands:
            if command in additional_commands:
                additional_commands.remove(command)

    if additional_commands:
        ctx.custom_commands = list(additional_commands)
        try:
            cmd_capture_plugin = CmdCapturePlugin(ctx)
            cmd_capture_plugin.run()
        except PluginError as e:
            ctx.warning("Failed to capture output of a command. Error: {}".format(e))

        ctx.custom_commands = user_defined_custom_commands

    return



# =============================================================================
# pre_migrate.py - plugin for preparing for migrating classic XR to eXR/fleXR
#
# Copyright (c)  2013, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import csv
import os
import re
import subprocess

import pexpect

from csmpe.plugins import CSMPlugin
from csmpe.core_plugins.csm_install_operations.utils import ServerType, is_empty, concatenate_dirs
from simple_server_helper import TFTPServer, FTPServer, SFTPServer
from hardware_audit import Plugin as HardwareAuditPlugin
from migration_lib import log_and_post_status
from csmpe.core_plugins.csm_get_inventory.ios_xr.plugin import get_package, get_inventory

MINIMUM_RELEASE_VERSION_FOR_MIGRATION = "6.1.3"

NOX_FOR_MAC = "nox-mac64.bin"
NOX_64_BINARY = "nox-linux-64.bin"

TIMEOUT_FOR_COPY_CONFIG = 3600
TIMEOUT_FOR_COPY_IMAGE = 1800
TIMEOUT_FOR_FPD_UPGRADE = 5400

IMAGE_LOCATION = "harddisk:/"
CONFIG_LOCATION = "harddiskb:/"

XR_CONFIG_IN_CSM = "xr.cfg"
ADMIN_CONFIG_IN_CSM = "admin.cfg"

CONVERTED_XR_CONFIG_IN_CSM = "xr.iox"
CONVERTED_ADMIN_CAL_CONFIG_IN_CSM = "admin.cal"
CONVERTED_ADMIN_XR_CONFIG_IN_CSM = "admin.iox"

FINAL_CAL_CONFIG = "cXR_admin_plane_converted_eXR.cfg"
FINAL_XR_CONFIG = "cXR_xr_plane_converted_eXR.cfg"

# XR_CONFIG_ON_DEVICE = "iosxr.cfg"
# ADMIN_CAL_CONFIG_ON_DEVICE = "admin_calvados.cfg"
# ADMIN_XR_CONFIG_ON_DEVICE = "admin_iosxr.cfg"


class Plugin(CSMPlugin):
    """
    A plugin for preparing device for migration from
    ASR9K IOS-XR (a.k.a. XR) to ASR9K IOS-XR 64 bit (a.k.a. eXR)

    This plugin does the following:
    1. Check several pre-requisites
    2. Resize the eUSB partition(/harddiskb:/ on XR)
    3. Migrate the configurations with NoX and upload them to device
    4. Copy the eXR image to /harddiskb:/
    5. Upgrade some FPD's if needed.

    Console access is needed.
    """
    name = "Pre-Migrate Plugin"
    platforms = {'ASR9K'}
    phases = {'Pre-Migrate'}
    os = {'XR'}

    node_pattern = re.compile("^\d+(/\w+)+$")

    def _save_show_platform(self):
        """Save the output of 'show platform' to session log"""

        cmd = "show platform"
        # show platform can take more than 1 minute after router reload. Issue No. 47
        output = self.ctx.send(cmd, timeout=600)
        file_name = self.ctx.save_to_file(cmd, output)
        if file_name is None:
            self.ctx.warning("Unable to save '{}' output to file: {}".format(cmd, file_name))

    def _ping_repository_check(self, repo_url):
        """Test ping server repository ip from device"""
        repo_ip = re.search("[/@](\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/", repo_url)

        if not repo_ip:
            self.ctx.error("Bad hostname for server repository. Please check the settings in CSM.")

        output = self.ctx.send("ping {}".format(repo_ip.group(1)))
        if "100 percent" not in output:
            self.ctx.error("Failed to ping server repository {} on device." +
                           "Please check session.log.".format(repo_ip.group(1)))

    def _all_configs_supported(self, nox_output):
        """Check text output from running NoX on system. Only return True if all configs are supported by eXR."""
        pattern = "Filename[\sA-Za-z\n]*[-\s]*\S*\s+\d*\s+\d*\(\s*\d*%\)\s+\d*\(\s*\d*%\)\s+\d*\(\s*\d*%\)\s+(\d*)"
        match = re.search(pattern, nox_output)

        if match:
            if match.group(1) != "0":
                return False

        return True

    def _upload_files_to_server_repository(self, sourcefiles, server, destfilenames):
        """
        Upload files from their locations in the host linux system to the FTP/TFTP/SFTP server repository.

        Arguments:
        :param sourcefiles: a list of string file paths that each points to a file on the system where CSM is hosted.
                            The paths are all relative to csm/csmserver/.
                            For example, if the source file is in csm_data/migration/filename,
                            sourcefiles = ["../../csm_data/migration/filename"]
        :param server: the associated server repository object stored in CSM database
        :param destfilenames: a list of string filenames that the source files should be named after being copied to
                              the designated directory in the server repository. i.e., ["thenewfilename"]
        :return: True if no error occurred.
        """

        server_type = server.server_type
        selected_server_directory = self.ctx._csm.install_job.server_directory
        if server_type == ServerType.TFTP_SERVER:
            tftp_server = TFTPServer(server)
            for x in range(0, len(sourcefiles)):
                log_and_post_status(self.ctx, "Copying file {} to {}/{}/{}.".format(sourcefiles[x],
                                                                                    server.server_directory,
                                                                                    selected_server_directory,
                                                                                    destfilenames[x]))
                try:
                    tftp_server.upload_file(sourcefiles[x], destfilenames[x],
                                            sub_directory=selected_server_directory)
                except:
                    self.ctx.error("Exception was thrown while " +
                                   "copying file {} to {}/{}/{}.".format(sourcefiles[x],
                                                                         server.server_directory,
                                                                         selected_server_directory,
                                                                         destfilenames[x]))

        elif server_type == ServerType.FTP_SERVER:
            ftp_server = FTPServer(server)
            for x in range(0, len(sourcefiles)):
                log_and_post_status(self.ctx, "Copying file {} to {}/{}/{}.".format(sourcefiles[x],
                                                                                    server.server_directory,
                                                                                    selected_server_directory,
                                                                                    destfilenames[x]))
                try:
                    ftp_server.upload_file(sourcefiles[x], destfilenames[x],
                                           sub_directory=selected_server_directory)
                except:
                    self.ctx.error("Exception was thrown while " +
                                   "copying file {} to {}/{}/{}.".format(sourcefiles[x],
                                                                         server.server_directory,
                                                                         selected_server_directory,
                                                                         destfilenames[x]))
        elif server_type == ServerType.SFTP_SERVER:
            sftp_server = SFTPServer(server)
            for x in range(0, len(sourcefiles)):
                log_and_post_status(self.ctx, "Copying file {} to {}/{}/{}.".format(sourcefiles[x],
                                                                                    server.server_directory,
                                                                                    selected_server_directory,
                                                                                    destfilenames[x]))
                try:
                    sftp_server.upload_file(sourcefiles[x], destfilenames[x],
                                            sub_directory=selected_server_directory)
                except:
                    self.ctx.error("Exception was thrown while " +
                                   "copying file {} to {}/{}/{}.".format(sourcefiles[x],
                                                                         server.server_directory,
                                                                         selected_server_directory,
                                                                         destfilenames[x]))
        else:
            self.ctx.error("Pre-Migrate does not support {} server repository.".format(server_type))

        return True

    def _copy_files_to_device(self, server, repository, source_filenames, dest_files, timeout=600):
        """
        Copy files from their locations in the user selected server directory in the FTP/TFTP/SFTP server repository
        to locations on device.

        Arguments:
        :param server: the server object fetched from database
        :param repository: the string url link that points to the location of files in the SFTP server repository
        :param source_filenames: a list of string filenames in the designated directory in the server repository.
        :param dest_files: a list of string file paths that each points to a file to be created on device.
                    i.e., ["harddiskb:/asr9k-mini-x64.tar"]
        :param timeout: the timeout for the sftp copy operation on device. The default is 10 minutes.
        :return: None if no error occurred.
        """

        if server.server_type == ServerType.FTP_SERVER or server.server_type == ServerType.TFTP_SERVER:
            self._copy_files_from_ftp_tftp_to_device(repository, source_filenames, dest_files, timeout=timeout)

        elif server.server_type == ServerType.SFTP_SERVER:
            self._copy_files_from_sftp_to_device(server, source_filenames, dest_files, timeout=timeout)

        else:
            self.ctx.error("Pre-Migrate does not support {} server repository.".format(server.server_type))

    def _copy_files_from_ftp_tftp_to_device(self, repository, source_filenames, dest_files, timeout=600):
        """
        Copy files from their locations in the user selected server directory in the FTP or TFTP server repository
        to locations on device.

        Arguments:
        :param repository: the string url link that points to the location of files in the FTP/TFTP server repository,
                    with no extra '/' in the end. i.e., tftp://223.255.254.245/tftpboot
        :param source_filenames: a list of string filenames in the designated directory in the server repository.
        :param dest_files: a list of string file paths that each points to a file to be created on device.
                    i.e., ["harddiskb:/asr9k-mini-x64.tar"]
        :param timeout: the timeout for the 'copy' CLI command on device. The default is 10 minutes.
        :return: None if no error occurred.
        """

        def send_newline(ctx):
            ctx.ctrl.sendline()
            return True

        def error(ctx):
            ctx.message = "Error copying file."
            return False

        for x in range(0, len(source_filenames)):

            command = "copy {}/{} {}".format(repository, source_filenames[x], dest_files[x])

            CONFIRM_FILENAME = re.compile("Destination filename.*\?")
            CONFIRM_OVERWRITE = re.compile("Copy : Destination exists, overwrite \?\[confirm\]")
            COPIED = re.compile(".+bytes copied in.+ sec")
            COPYING = re.compile("C" * 50)
            NO_SUCH_FILE = re.compile("%Error copying.*\(Error opening source file\): No such file or directory")
            ERROR_COPYING = re.compile("%Error copying")

            PROMPT = self.ctx.prompt
            TIMEOUT = self.ctx.TIMEOUT

            events = [PROMPT, CONFIRM_FILENAME, CONFIRM_OVERWRITE, COPIED, COPYING,
                      TIMEOUT, NO_SUCH_FILE, ERROR_COPYING]
            transitions = [
                (CONFIRM_FILENAME, [0], 1, send_newline, timeout),
                (CONFIRM_OVERWRITE, [1], 2, send_newline, timeout),
                (COPIED, [1, 2], 3, None, 20),
                (COPYING, [1, 2], 2, send_newline, timeout),
                (PROMPT, [3], -1, None, 0),
                (TIMEOUT, [0, 1, 2, 3], -1, error, 0),
                (NO_SUCH_FILE, [0, 1, 2, 3], -1, error, 0),
                (ERROR_COPYING, [0, 1, 2, 3], -1, error, 0),
            ]

            log_and_post_status(self.ctx, "Copying {}/{} to {} on device".format(repository,
                                                                                 source_filenames[x],
                                                                                 dest_files[x]))

            if not self.ctx.run_fsm("Copy file from tftp/ftp to device", command, events, transitions,
                                    timeout=20, max_transitions=40):
                self.ctx.error("Error copying {}/{} to {} on device".format(repository,
                                                                            source_filenames[x],
                                                                            dest_files[x]))

            output = self.ctx.send("dir {}".format(dest_files[x]))
            if "No such file" in output:
                self.ctx.error("Failed to copy {}/{} to {} on device".format(repository,
                                                                             source_filenames[x],
                                                                             dest_files[x]))

    def _copy_files_from_sftp_to_device(self, server, source_filenames, dest_files, timeout=600):
        """
        Copy files from their locations in the user selected server directory in the SFTP server repository
        to locations on device.

        Arguments:
        :param server: the sftp server object
        :param source_filenames: a list of string filenames in the designated directory in the server repository.
        :param dest_files: a list of string file paths that each points to a file to be created on device.
                    i.e., ["harddiskb:/asr9k-mini-x64.tar"]
        :param timeout: the timeout for the sftp copy operation on device. The default is 10 minutes.
        :return: None if no error occurred.
        """
        source_path = server.server_url

        remote_directory = concatenate_dirs(server.server_directory, self.ctx._csm.install_job.server_directory)
        if not is_empty(remote_directory):
            source_path += ":{}".format(remote_directory)

        def pause_logfile_stream(ctx):
            """
            This was made necessary because during sftp download, when file is large,
            the number of transferred bytes keeps changing and session log takes so much
            time in reading and writing the changing number that it is still doing that
            long after the operation is complete.
            """
            if ctx.ctrl._session.logfile_read:
                ctx.ctrl._session.logfile_read = None
            return True

        def send_password(ctx):
            ctx.ctrl.sendline(server.password)
            return pause_logfile_stream(ctx)

        def send_yes(ctx):
            ctx.ctrl.sendline("yes")
            return pause_logfile_stream(ctx)

        def reinstall_logfile(ctx):
            if ctx.ctrl._logfile_fd and (not ctx.ctrl._session.logfile_read):
                ctx.ctrl._session.logfile_read = ctx.ctrl._logfile_fd
            else:
                ctx.message = "Error reinstalling session.log."
                return False
            return True

        def timeout_error(ctx):
            reinstall_logfile(ctx)
            ctx.message = "Timed out while copying file from sftp."
            return False

        def no_such_file_error(ctx):
            reinstall_logfile(ctx)
            ctx.message = "Copying the file from sftp failed because it is not found in the specified path."
            return False

        def download_abort_error(ctx):
            reinstall_logfile(ctx)
            ctx.message = "Copying the file from sftp failed. Download was aborted."
            return False

        for x in range(0, len(source_filenames)):
            if is_empty(server.vrf):
                command = "sftp {}@{}/{} {}".format(server.username, source_path, source_filenames[x], dest_files[x])
            else:
                command = "sftp {}@{}/{} {} vrf {}".format(server.username, source_path, source_filenames[x],
                                                           dest_files[x], server.vrf)

            PASSWORD = re.compile("Password:")
            CONFIRM_OVERWRITE = re.compile("Overwrite.*\[yes/no\]\:")
            COPIED = re.compile("bytes copied in", re.MULTILINE)
            NO_SUCH_FILE = re.compile("src.*does not exist")
            DOWNLOAD_ABORTED = re.compile("Download aborted.")

            PROMPT = self.ctx.prompt
            TIMEOUT = self.ctx.TIMEOUT

            events = [PROMPT, PASSWORD, CONFIRM_OVERWRITE, COPIED, TIMEOUT, NO_SUCH_FILE, DOWNLOAD_ABORTED]
            transitions = [
                (PASSWORD, [0], 1, send_password, timeout),
                (CONFIRM_OVERWRITE, [1], 2, send_yes, timeout),
                (COPIED, [1, 2], -1, reinstall_logfile, 0),
                (PROMPT, [1, 2], -1, reinstall_logfile, 0),
                (TIMEOUT, [0, 1, 2], -1, timeout_error, 0),
                (NO_SUCH_FILE, [0, 1, 2], -1, no_such_file_error, 0),
                (DOWNLOAD_ABORTED, [0, 1, 2], -1, download_abort_error, 0),
            ]

            log_and_post_status(self.ctx, "Copying {}/{} to {} on device".format(source_path,
                                                                                 source_filenames[x],
                                                                                 dest_files[x]))

            if not self.ctx.run_fsm("Copy file from sftp to device", command, events, transitions, timeout=20):
                self.ctx.error("Error copying {}/{} to {} on device".format(source_path,
                                                                            source_filenames[x],
                                                                            dest_files[x]))

            output = self.ctx.send("dir {}".format(dest_files[x]))
            if "No such file" in output:
                self.ctx.error("Failed to copy {}/{} to {} on device".format(source_path,
                                                                             source_filenames[x],
                                                                             dest_files[x]))

    def _run_migration_on_config(self, fileloc, filename, nox_to_use, hostname):
        """
        Run the migration tool - NoX - on the configurations copied out from device.

        The conversion/migration is successful if the number under 'Total' equals to
        the number under 'Known' in the text output.

        If it's successful, but not all existing configs are supported by eXR, create two
        new log files for the supported and unsupported configs in session log directory.
        The unsupported configs will not appear on the converted configuration files.
        Log a warning about the removal of unsupported configs, but this is not considered
        as error.

        If it's not successful, meaning that there are some configurations not known to
        the NoX tool, in this case, create two new log files for the supported and unsupported
        configs in session log directory. After that, error out.

        :param fileloc: string location where the config needs to be converted/migrated is,
                        without the '/' in the end. This location is relative to csm/csmserver/
        :param filename: string filename of the config
        :param nox_to_use: string name of NoX binary executable.
        :param hostname: hostname of device, as recorded on CSM.
        :return: None if no error occurred.
        """

        try:
            commands = [subprocess.Popen(["chmod", "+x", nox_to_use]),
                        subprocess.Popen([nox_to_use, "-f", os.path.join(fileloc, filename)],
                                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                        ]

            nox_output, nox_error = commands[1].communicate()
        except OSError:
            self.ctx.error("Failed to run the configuration migration tool {} on config file {} - OSError.".format(
                nox_to_use,
                os.path.join(fileloc, filename))
            )

        if nox_error:
            self.ctx.error("Failed to run the configuration migration tool on the admin configuration " +
                           "we retrieved from device - {}.".format(nox_error))

        if filename.split('.')[0] == 'admin':
            if (not os.path.isfile(os.path.join(fileloc, CONVERTED_ADMIN_CAL_CONFIG_IN_CSM))) or \
               (not os.path.isfile(os.path.join(fileloc, CONVERTED_ADMIN_XR_CONFIG_IN_CSM))):
                self.ctx.error("Failed to convert the ASR9K admin configuration with NoX tool.")

        elif not os.path.isfile(os.path.join(fileloc, CONVERTED_XR_CONFIG_IN_CSM)):
            self.ctx.error("Failed to convert the ASR9K IOS-XR configuration with NoX tool.")

        conversion_success = False

        match = re.search("Filename[\sA-Za-z\n]*[-\s]*\S*\s+(\d*)\s+\d*\(\s*\d*%\)\s+\d*\(\s*\d*%\)\s+(\d*)",
                          nox_output)

        if match:
            if match.group(1) == match.group(2):
                conversion_success = True

        if filename == ADMIN_CONFIG_IN_CSM:
            supported_log_name = "supported_config_in_admin_configuration"
            unsupported_log_name = "unsupported_config_in_admin_configuration"
        else:
            supported_log_name = "supported_config_in_xr_configuration"
            unsupported_log_name = "unsupported_config_in_xr_configuration"

        if conversion_success:

            if self._all_configs_supported(nox_output):
                log_and_post_status(self.ctx, "Configuration {} was migrated successfully. ".format(filename) +
                                    "No unsupported configurations found.")
            else:
                self._create_config_logs(os.path.join(fileloc, filename.split(".")[0] + ".csv"),
                                         supported_log_name, unsupported_log_name,
                                         hostname, filename)

                log_and_post_status(self.ctx, "Configurations that are unsupported in eXR were removed in " +
                                    "{}. Please look into {} and {}.".format(filename,
                                                                             unsupported_log_name,
                                                                             supported_log_name))
        else:
            self._create_config_logs(os.path.join(fileloc, filename.split(".")[0] + ".csv"),
                                     supported_log_name, unsupported_log_name, hostname, filename)

            self.ctx.error("Unknown configurations found. Please look into {} ".format(unsupported_log_name) +
                           "for unprocessed configurations, and {} for known/supported configurations".format(
                           unsupported_log_name, supported_log_name)
                           )

    def _resize_eusb(self):
        """Resize the eUSB partition on device - Run the /pkg/bin/resize_eusb script on device(from ksh)."""

        output = self.ctx.send("run /pkg/bin/resize_eusb")

        if "Pre-Migration Operation Completed." not in output:
            self.ctx.error("Pre-Migrate partition check failed. Please check session.log.")
        # output = self.ctx.send("show media")

        # eusb_size = re.search("/harddiskb:.* ([.\d]+)G", output)

        # if not eusb_size:
        #    self.ctx.error("Unexpected output from CLI 'show media'.")

        # if eusb_size.group(1) < "1.0":
        #    self.ctx.error("/harddiskb:/ is smaller than 1 GB after running /pkg/bin/resize_eusb. " +
        #                   "Please make sure that the device has either RP2 or RSP4.")

    def _check_fpd(self, fpd_relevant_nodes):
        """
        Check the versions of migration related FPD's on device. Return a dictionary
        that tells which FPD's on which nodes require successful FPD upgrade later on.

        :param fpd_relevant_nodes: a dictionary. Keys are strings representing all node locations
                                   on device parsed from output of "admin show platform".
                                   Values are integers. Value can either be 0 or 1.
                                   value 1 means that we actually will need to make sure that the
                                   FPD upgrade later on for this node location completes successfully,
                                   value 0 means that we don't need to check if the
                                   FPD upgrade later on for this node location is successful or not.
        :return: a dictionary with string FPD type as key, and a set of the string names of
                 node locations as value.
        """
        fpdtable = self.ctx.send("show hw-module fpd location all")

        subtype_to_locations_need_upgrade = {}

        last_location = None
        for line in fpdtable.split('\n'):

            first_word = line.split(' ', 1)[0]

            if self.node_pattern.match(first_word):
                # since fpd_relevant_nodes is loaded from db, the keys are
                # unicode instead of byte strings
                indicator = fpd_relevant_nodes.get(unicode(first_word, encoding="latin1"))
                # indicator is 1:
                #       Detect a new node(RSP/RP/LC/FC) of which fpds we'll need to check
                #       if upgrade goes successful
                # indicator is None:
                #       Detect node that is not found in output of "admin show platform"
                #       we need to check if FPD upgrade goes successful in this case
                if indicator == 1 or indicator is None:
                    last_location = first_word
                # indicator is 0:
                #       Detect node to be PEM/FAN or some other unsupported hardware in eXR.
                #       we don't care if the FPD upgrade for these is successful or not
                #       so we update last_location to None
                else:
                    last_location = None

            # Found some fpd that needs upgrade
            if last_location and len(line) >= 79 and line[76:79] == "Yes":
                fpdtype_end_idx = 51
                while line[fpdtype_end_idx] != ' ':
                    fpdtype_end_idx += 1

                fpdtype = line[51:fpdtype_end_idx]

                if fpdtype not in subtype_to_locations_need_upgrade:
                    # it is possible to have duplicates, so using set here
                    subtype_to_locations_need_upgrade[fpdtype] = set()
                subtype_to_locations_need_upgrade[fpdtype].add(last_location)

        return subtype_to_locations_need_upgrade

    def _check_if_fpd_package_installed(self):
        """
        Check if the FPD package is already active on device.
        Error out if not.

        :return: None if FPD package is active, error out if not.
        """
        active_packages = self.ctx.send("show install active summary")

        match = re.search("fpd", active_packages)

        if not match:
            self.ctx.error("No FPD package is active on device. Please install the FPD package on device first.")

        return

    def _ensure_updated_fpd(self, fpd_relevant_nodes):
        """
        Upgrade FPD's if needed.
        Steps:
        1. Check version of the migration related FPD's. Get the dictionary
           of FPD types mapped to locations for which we need to check for
           upgrade successs.
        2. Force install the FPD types that need upgrade on all locations.
           Check FPD related sys log to make sure all necessary upgrades
           defined by the dictionary complete successfully.

        :param fpd_relevant_nodes: a dictionary. Keys are strings representing all node locations
                                   on device parsed from output of "admin show platform".
                                   Values are integers. Value can either be 0 or 1.
                                   value 1 means that we actually will need to make sure that the
                                   FPD upgrade later on for this node location completes successfully,
                                   value 0 means that we don't need to check if the
                                   FPD upgrade later on for this node location is successful or not.

        :return: True if no error occurred.
        """
        # check for the FPD version, if FPD needs upgrade,
        log_and_post_status(self.ctx, "Checking FPD versions...")
        subtype_to_locations_need_upgrade = self._check_fpd(fpd_relevant_nodes)

        if subtype_to_locations_need_upgrade:

            # Force upgrade all FPD's in RP and Line card that need upgrade, with the FPD pie or both the FPD
            # pie and FPD SMU depending on release version
            self._upgrade_all_fpds(subtype_to_locations_need_upgrade)

        return True

    def _upgrade_all_fpds(self, subtype_to_locations_need_upgrade):
        """Force upgrade certain FPD's on all locations. Check for success. """
        def send_newline(ctx):
            ctx.ctrl.sendline()
            return True

        def send_yes(ctx):
            ctx.ctrl.sendline("yes")
            return True

        def error(ctx):
            ctx.message = "Error upgrading FPD."
            return False

        def timeout(ctx):
            ctx.message = "Timeout upgrading FPD."
            return False

        for fpdtype in subtype_to_locations_need_upgrade:

            log_and_post_status(self.ctx, "FPD upgrade - start to upgrade FPD {} on all locations".format(fpdtype))

            CONFIRM_CONTINUE = re.compile("Continue\? \[confirm\]")
            CONFIRM_SECOND_TIME = re.compile("Continue \? \[no\]:")
            UPGRADE_END = re.compile("FPD upgrade has ended.")

            PROMPT = self.ctx.prompt
            TIMEOUT = self.ctx.TIMEOUT

            events = [PROMPT, CONFIRM_CONTINUE, CONFIRM_SECOND_TIME, UPGRADE_END, TIMEOUT]
            transitions = [
                (CONFIRM_CONTINUE, [0], 1, send_newline, TIMEOUT_FOR_FPD_UPGRADE),
                (CONFIRM_SECOND_TIME, [0, 1], 2, send_yes, TIMEOUT_FOR_FPD_UPGRADE),
                (UPGRADE_END, [1, 2], 3, None, 120),
                (PROMPT, [3], -1, None, 0),
                (PROMPT, [1, 2], -1, error, 0),
                (TIMEOUT, [0, 1, 2], -1, timeout, 0),

            ]

            if not self.ctx.run_fsm("Upgrade FPD",
                                    "admin upgrade hw-module fpd {} location all".format(fpdtype),
                                    events, transitions, timeout=30):
                self.ctx.error("Error while upgrading FPD subtype {}. Please check session.log".format(fpdtype))

            fpd_log = self.ctx.send("show log | include fpd", timeout=300)

            for location in subtype_to_locations_need_upgrade[fpdtype]:

                pattern = "Successfully\s*(?:downgrade|upgrade)\s*{}.*location\s*{}".format(fpdtype, location)
                fpd_upgrade_success = re.search(pattern, fpd_log)

                if not fpd_upgrade_success:
                    self.ctx.error("Failed to upgrade FPD subtype {} on location {}. ".format(fpdtype, location) +
                                   "Please check session.log.")
        return True

    def _create_config_logs(self, csvfile, supported_log_name, unsupported_log_name, hostname, filename):
        """
        Create two logs for migrated configs that are unsupported and supported by eXR.
        They are stored in the same directory as session log, for user to view.

        :param csvfile: the string csv filename generated by running NoX on original config.
        :param supported_log_name: the string filename for the supported configs log
        :param unsupported_log_name: the string filename for the unsupported configs log
        :param hostname: string hostname of device, as recorded on CSM.
        :param filename: string filename of original config
        :return: None if no error occurred
        """

        if not os.path.isfile(os.path.join(csvfile)):
            self.ctx.error("Missing the csv file {} that should have been generated by the NoX tool".format(csvfile) +
                           " during the configuration conversion. Failed to write diagnostic files.")
        supported_config_log = os.path.join(self.ctx.log_directory, supported_log_name)
        unsupported_config_log = os.path.join(self.ctx.log_directory, unsupported_log_name)
        try:
            with open(supported_config_log, 'w') as supp_log:
                with open(unsupported_config_log, 'w') as unsupp_log:
                    supp_log.write('Configurations Known and Supported to the NoX Conversion Tool \n \n')

                    unsupp_log.write('Configurations Unprocessed by the NoX Conversion Tool (Comments, Markers,' +
                                     ' or Unknown/Unsupported Configurations) \n \n')

                    supp_log.write('{0[0]:<8} {0[1]:^20} \n'.format(("Line No.", "Configuration")))
                    unsupp_log.write('{0[0]:<8} {0[1]:^20} \n'.format(("Line No.", "Configuration")))
                    with open(csvfile, 'rb') as csvfile:
                        reader = csv.reader(csvfile)
                        for row in reader:
                            if len(row) >= 3 and row[1].strip() == "KNOWN_SUPPORTED":
                                supp_log.write('{0[0]:<8} {0[1]:<} \n'.format((row[0], row[2])))
                            elif len(row) >= 3:
                                unsupp_log.write('{0[0]:<8} {0[1]:<} \n'.format((row[0], row[2])))

                    msg = "\n \nPlease find original configuration in csm_data/migration/{}/{} \n".format(hostname,
                                                                                                          filename)
                    supp_log.write(msg)
                    unsupp_log.write(msg)
                    if filename.split('.')[0] == 'admin':
                        msg2 = "The final converted configuration is in csm_data/migration/" + \
                               hostname + "/" + CONVERTED_ADMIN_CAL_CONFIG_IN_CSM + \
                               " and csm_data/migration/" + hostname + "/" + CONVERTED_ADMIN_XR_CONFIG_IN_CSM
                    else:
                        msg2 = "The final converted configuration is in csm_data/migration/" + \
                               hostname + "/" + CONVERTED_XR_CONFIG_IN_CSM
                    supp_log.write(msg2)
                    unsupp_log.write(msg2)
                    csvfile.close()
                unsupp_log.close()
            supp_log.close()
        except:
            self.ctx.error("Error writing diagnostic files - in " + self.ctx.log_directory +
                           " during configuration migration.")

    def _filter_server_repository(self, server):
        """Filter out LOCAL server repositories and only keep TFTP, FTP and SFTP"""
        if not server:
            self.ctx.error("Pre-Migrate missing server repository object.")
        if server.server_type != ServerType.FTP_SERVER and server.server_type != ServerType.TFTP_SERVER and \
           server.server_type != ServerType.SFTP_SERVER:
            self.ctx.error("Pre-Migrate does not support " + server.server_type + " server repository.")

    def _save_config_to_csm_data(self, files, admin=False):
        """
        Copy the admin configuration or IOS-XR configuration
        from device to csm_data.

        :param files: the full local file paths for configs.
        :param admin: True if asking for admin config, False otherwise.
        :return: None
        """

        try:
            cmd = "admin show run" if admin else "show run"
            output = self.ctx.send(cmd, timeout=TIMEOUT_FOR_COPY_CONFIG)
            ind = output.rfind('Building configuration...\n')

        except pexpect.TIMEOUT:
            self.ctx.error("CLI '{}' timed out after 1 hour.".format(cmd))

        for file_path in files:
            # file = '../../csm_data/migration/<hostname>' + filename
            file_to_write = open(file_path, 'w+')
            file_to_write.write(output[(ind + 1):])
            file_to_write.close()

    def _handle_configs(self, hostname, server, repo_url, fileloc, nox_to_use, config_filename):
        """
        1. Copy admin and XR configs from device to TFTP/FTP/SFTP server repository.
        2. Copy admin and XR configs from server repository to csm_data/migration/<hostname>/
        3. Copy admin and XR configs from server repository to session log directory as
           show-running-config.txt and admin-show-running-config.txt for comparisons
           after Migrate or Post-Migrate. (Diff will be generated.)
        4. Run NoX on admin config first. This run generates 1) eXR admin/calvados config
           and POSSIBLY 2) eXR XR config.
        5. Run NoX on XR config if no custom eXR config has been selected by user when
           Pre-Migrate is scheduled. This run generates eXR XR config.
        6. Merge either the selected eXR custom config or the converted XR config with the converted
           eXR XR config to form a new file - cXR_xr_plane_converted_eXR.cfg
        7. Copy the eXR admin/calvados config and the cXR_xr_plane_converted_eXR.cfg to the server
           repository and then from there to device.
           Note if user selected custom eXR XR config, that will be uploaded instead of
           the NoX migrated original XR config.

        :param hostname: string hostname of device, as recorded on CSM.
        :param repo_url: the URL of the selected server repository. i.e., tftp://223.255.254.245/tftpboot
        :param fileloc: the string path ../../csm_data/migration/<hostname>
        :param nox_to_use: the name of the NoX binary executable
        :param config_filename: the user selected string filename of custom eXR XR config.
                                If it's '', nothing was selected.
                                If selected, this file must be in the server repository.
        :return: None if no error occurred.
        """

        log_and_post_status(self.ctx, "Cleaning up previously saved configuration files for this host in csm_data")
        for old_file in os.listdir(fileloc):
            try:
                os.remove(os.path.join(fileloc, old_file))
            except:
                self.ctx.warning("Failed to remove the old configuration conversion file " +
                                 "{}".format(os.path.join(fileloc, old_file)))
                pass

        log_and_post_status(self.ctx, "Saving the current configurations on device into server repository and csm_data")

        self._save_config_to_csm_data([os.path.join(fileloc, ADMIN_CONFIG_IN_CSM),
                                       os.path.join(self.ctx.log_directory,
                                       self.ctx.normalize_filename("admin show running-config"))
                                       ], admin=True)

        self._save_config_to_csm_data([os.path.join(fileloc, XR_CONFIG_IN_CSM),
                                       os.path.join(self.ctx.log_directory,
                                       self.ctx.normalize_filename("show running-config"))
                                       ], admin=False)

        log_and_post_status(self.ctx, "Converting admin configuration file with configuration migration tool")
        self._run_migration_on_config(fileloc, ADMIN_CONFIG_IN_CSM, nox_to_use, hostname)

        # ["admin.cal"]
        config_files = [CONVERTED_ADMIN_CAL_CONFIG_IN_CSM]
        # ["cXR_admin_plane_converted_eXR.cfg"]
        config_names_on_device = [FINAL_CAL_CONFIG]
        if not config_filename:

            log_and_post_status(self.ctx, "Converting IOS-XR configuration file with configuration migration tool")
            self._run_migration_on_config(fileloc, XR_CONFIG_IN_CSM, nox_to_use, hostname)

            # admin.iox and xr.iox
            files_to_merge = [os.path.join(fileloc, CONVERTED_ADMIN_XR_CONFIG_IN_CSM),
                              os.path.join(fileloc, CONVERTED_XR_CONFIG_IN_CSM)]

            with open(os.path.join(fileloc, FINAL_XR_CONFIG), 'w') as merged_file:
                for fname in files_to_merge:
                    with open(fname) as infile:
                        for line in infile:
                            merged_file.write(line)

            # "cXR_xr_plane_converted_eXR.cfg" - product of files_to_merge, merging will be done
            config_files.append(FINAL_XR_CONFIG)
            # "cXR_xr_plane_converted_eXR.cfg"
            config_names_on_device.append(FINAL_XR_CONFIG)

        log_and_post_status(self.ctx, "Uploading the migrated configuration files to server repository and device.")

        config_names_in_repo = [hostname + "_" + config_name for config_name in config_files]

        if self._upload_files_to_server_repository([os.path.join(fileloc, config_name)
                                                    for config_name in config_files],
                                                   server, config_names_in_repo):

            if config_filename:
                config_names_in_repo.append(config_filename)
                # "cXR_xr_plane_converted_eXR.cfg"
                config_names_on_device.append(FINAL_XR_CONFIG)

            self._copy_files_to_device(server, repo_url, config_names_in_repo,
                                       [CONFIG_LOCATION + config_name
                                        for config_name in config_names_on_device],
                                       timeout=TIMEOUT_FOR_COPY_CONFIG)

    def _get_packages(self, packages):
        """Find out which package is eXR tar file, which is crypto_auto_key_gen.txt"""
        if len(packages) > 2:
            self.ctx.error("More than two packages are selected, however, only the ASR9K IOS XR 64 Bit tar file and the crypto key generation file should be selected.")
        if len(packages) == 0:
            self.ctx.error("No ASR9K IOS XR 64 Bit tar file selected for Pre-Migrate.")
        image_pattern = re.compile("asr9k.*\.tar.*")
        exr_tar = None
        crypto_file = None
        for package in packages:
            if image_pattern.match(package):
                exr_tar = package
            else:
                crypto_file = package
        return exr_tar, crypto_file

    def _find_nox_to_use(self):
        """
        Find out if the linux system is 32 bit or 64 bit. NoX currently only has a binary executable
        compiled for 64 bit.
        """
        check_32_or_64_system = subprocess.Popen(['uname', '-a'], stdout=subprocess.PIPE)

        out, err = check_32_or_64_system.communicate()

        if err:
            self.ctx.error("Failed to execute 'uname -a' on the linux system.")

        if "x86_64" in out:
            return NOX_64_BINARY
        else:
            self.ctx.error("The configuration migration tool NoX is currently not available for 32 bit linux system.")

    def run(self):
        server_repo_url = None
        try:
            server_repo_url = self.ctx.server_repository_url
        except AttributeError:
            pass

        if server_repo_url is None:
            self.ctx.error("No repository provided.")

        try:
            packages = self.ctx.software_packages
        except AttributeError:
            self.ctx.error("No package list provided")

        config_filename_tuple = self.ctx.load_job_data('config_filename')
        if config_filename_tuple:
            config_filename = config_filename_tuple[0]

        server = None
        try:
            server = self.ctx.get_server
        except AttributeError:
            self.ctx.error("No server repository selected")

        if server is None:
            self.ctx.error("No server repository selected")

        if not self.ctx.load_job_data('override_hw_req'):
            self.ctx.error("Missing indication of whether to override hardware requirement or not.")

        exr_image, crypto_file = self._get_packages(packages)

        version_match = re.findall("(\d+\.\d+)\.\d+", exr_image)
        if version_match:
            exr_version = version_match[0]
        else:
            self.ctx.error("The selected tar file is missing release number in its filename.")

        self._filter_server_repository(server)

        hostname_for_filename = re.sub("[()\s]", "_", self.ctx._csm.host.hostname)
        hostname_for_filename = re.sub("_+", "_", hostname_for_filename)

        fileloc = self.ctx.migration_directory + hostname_for_filename

        if not os.path.exists(fileloc):
            os.makedirs(fileloc)

        self.ctx.save_job_data('hardware_audit_version', exr_version)
        hardware_audit_plugin = HardwareAuditPlugin(self.ctx)
        hardware_audit_plugin.run()

        fpd_relevant_nodes_tuple = self.ctx.load_job_data('fpd_relevant_nodes')
        if fpd_relevant_nodes_tuple:
            fpd_relevant_nodes = fpd_relevant_nodes_tuple[0]
        else:
            self.ctx.error("No data field fpd_relevant_nodes after completing hardware audit successfully.")

        log_and_post_status(self.ctx, "Checking current software version.")

        match_version = re.search("(\d\.\d\.\d).*", self.ctx.os_version)

        if not match_version:
            self.ctx.error("Bad os_version.")

        version = match_version.group(1)

        if version < MINIMUM_RELEASE_VERSION_FOR_MIGRATION:
            self.ctx.error("The minimal release version required for migration is {0}. Please upgrade to at lease {0} before scheduling migration.".format(MINIMUM_RELEASE_VERSION_FOR_MIGRATION))

        log_and_post_status(self.ctx, "Testing ping to selected server repository IP.")
        self._ping_repository_check(server_repo_url)

        log_and_post_status(self.ctx, "Checking if FPD package is active on device.")
        self._check_if_fpd_package_installed()

        nox_to_use = self.ctx.migration_directory + self._find_nox_to_use()

        if not os.path.isfile(nox_to_use):
            self.ctx.error("The configuration conversion tool {} is missing. ".format(nox_to_use) +
                           "CSM should have downloaded it from CCO when migration actions were scheduled.")

        self._save_show_platform()

        log_and_post_status(self.ctx, "Partition check and disk clean-up.")
        self._resize_eusb()

        self._handle_configs(hostname_for_filename, server,
                             server_repo_url, fileloc, nox_to_use, config_filename)

        log_and_post_status(self.ctx, "Copying the ASR9K-X64 image from server repository to device.")
        self._copy_files_to_device(server, server_repo_url, [exr_image],
                                   [IMAGE_LOCATION + exr_image], timeout=TIMEOUT_FOR_COPY_IMAGE)

        if crypto_file:
            log_and_post_status(self.ctx, "Copying the crypto key generation file from server repository to device.")
            self._copy_files_to_device(server, server_repo_url, [crypto_file],
                                       [CONFIG_LOCATION + crypto_file], timeout=60)

        self._ensure_updated_fpd(fpd_relevant_nodes)

        # Refresh package and inventory information
        get_package(self.ctx)
        get_inventory(self.ctx)

        return True



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================
import re
import time
import itertools
from condoor import ConnectionError, CommandError
from csmpe.core_plugins.csm_node_status_check.ios_xr.plugin_lib import parse_show_platform

install_error_pattern = re.compile("Error:    (.*)$", re.MULTILINE)

plugin_ctx = None


def send_yes(fsm_ctx):
    plugin_ctx.send('Y')
    return True


def log_install_errors(ctx, output):
    errors = re.findall(install_error_pattern, output)
    for line in errors:
        ctx.warning(line)


def watch_operation(ctx, op_id=0):
    """
    Function to keep watch on progress of operation
    and report KB downloaded.

    """
    no_install = r"There are no install requests in operation"
    op_progress = r"The operation is (\d+)% complete"
    op_download = r"(.*)KB downloaded: Download in progress"
    success = "Install operation {} completed successfully".format(op_id)

    cmd_show_install_request = "admin show install request"

    ctx.info("Watching the operation {} to complete".format(op_id))

    propeller = itertools.cycle(["|", "/", "-", "\\", "|", "/", "-", "\\"])

    output = None
    last_status = None
    finish = False
    time_tried = 0
    while not finish:
        try:
            try:
                # this is to catch the successful operation as soon as possible
                ctx.send("", wait_for_string=success, timeout=20)
                finish = True
            except ctx.CommandTimeoutError:
                pass

            message = ""
            # on CRS, it is observed that during Add, any command typed hangs for a while
            output = ctx.send(cmd_show_install_request, timeout=300)
            if op_id in output:
                # FIXME reconsider the logic here
                result = re.search(op_progress, output)
                if result:
                    status = result.group(0)
                    message = "{} {}".format(propeller.next(), status)

                result = re.search(op_download, output)
                if result:
                    status = result.group(0)
                    message += "\r\n<br>{}".format(status)

                if message != last_status:
                    ctx.post_status(message)
                    last_status = message
        except (ConnectionError, ctx.CommandTimeoutError) as e:
            if time_tried > 2:
                raise e

            time_tried += 1
            ctx.disconnect()
            time.sleep(60)
            ctx.reconnect()

        if no_install in output:
            break

    return output


def validate_node_state(inventory):
    valid_state = [
        'IOS XR RUN',
        'PRESENT',
        'UNPOWERED',
        'READY',
        'UNPOWERED',
        'FAILED',
        'OK',
        'ADMIN DOWN',
        'DISABLED'
    ]
    for key, value in inventory.items():
        if 'CPU' in key:
            if value['state'] not in valid_state:
                break
    else:
        return True
    return False


def wait_for_reload(ctx):
    """
     Wait for system to come up with max timeout as 25 Minutes

    """
    begin = time.time()
    if not ctx.is_console:
        ctx.disconnect()
        ctx.post_status("Waiting for device boot to reconnect")
        ctx.info("Waiting for device boot to reconnect")
        time.sleep(60)
        ctx.reconnect(max_timeout=3600, force_discovery=True)  # 60 * 60 = 3600

    else:
        ctx.info("Keeping console connected")
        ctx.post_status("Boot process started")
        ctx.info("Boot process started")
        if not ctx.reload(reload_timeout=1500, no_reload_cmd=True):
            ctx.error("Encountered error when attempting to reload device.")
        ctx.info("Boot process finished")

    ctx.info("Device connected successfully")

    timeout = 3600
    poll_time = 30
    time_waited = 0
    xr_run = "IOS XR RUN"

    cmd = "admin show platform"
    ctx.info("Waiting for all nodes to come up")
    ctx.post_status("Waiting for all nodes to come up")
    time.sleep(100)

    output = None

    while 1:
        # Wait till all nodes are in XR run state
        time_waited += poll_time
        if time_waited >= timeout:
            break

        time.sleep(poll_time)

        # show platform can take more than 1 minute after router reload. Issue No. 47
        output = ctx.send(cmd, timeout=600)
        if xr_run in output:
            inventory = parse_show_platform(ctx, output)
            if validate_node_state(inventory):
                ctx.info("All nodes in operational state")
                elapsed = time.time() - begin
                ctx.info("Overall outage time: {} minute(s) {:.0f} second(s)".format(elapsed // 60, elapsed % 60))
                return True

    # Some nodes did not come to run state
    ctx.error("Not all nodes have came up: {}".format(output))
    # this will never be executed
    return False


def watch_install(ctx, cmd, op_id=0):
    success_oper = r'Install operation (\d+) completed successfully'
    completed_with_failure = 'Install operation (\d+) completed with failure'
    failed_oper = r'Install operation (\d+) failed'
    failed_incr = r'incremental.*parallel'
    # restart = r'Parallel Process Restart'
    install_method = r'Install [M|m]ethod: (.*)'
    op_success = "The install operation will continue asynchronously"

    watch_operation(ctx, op_id)

    output = ctx.send("admin show install log {} detail".format(op_id))
    if re.search(failed_oper, output):
        if re.search(failed_incr, output):
            ctx.info("Retrying with parallel reload option")
            cmd += " parallel-reload"
            output = ctx.send(cmd)
            if op_success in output:
                result = re.search('Install operation (\d+) \'', output)
                if result:
                    op_id = result.group(1)
                    watch_operation(ctx, op_id)
                    output = ctx.send("admin show install log {} detail".format(op_id))
                else:
                    log_install_errors(ctx, output)
                    ctx.error("Operation ID not found")
                    return
        else:
            log_install_errors(ctx, output)
            ctx.error(output)
            return

    result = re.search(install_method, output)
    if result:
        restart_type = result.group(1).strip()
        ctx.info("{} Pending".format(restart_type))
        if restart_type == "Parallel Reload":
            if re.search(completed_with_failure, output):
                ctx.info("Install completed with failure, going for reload")
            elif re.search(success_oper, output):
                ctx.info("Install completed successfully, going for reload")
            return wait_for_reload(ctx)
        elif restart_type == "Parallel Process Restart":
            return True

    log_install_errors(ctx, output)
    return False


def install_add_remove(ctx, cmd, has_tar=False):
    message = "Waiting the operation to continue asynchronously"
    ctx.info(message)
    ctx.post_status(message)

    output = ctx.send(cmd, timeout=7200)
    result = re.search('Install operation (\d+) \'', output)
    if result:
        op_id = result.group(1)
    else:
        log_install_errors(ctx, output)
        ctx.error("Operation failed")
        return  # for sake of clarity

    op_success = "The install operation will continue asynchronously"
    failed_oper = r'Install operation {} failed'.format(op_id)
    if op_success in output:
        watch_operation(ctx, op_id=op_id)
        output = ctx.send("admin show install log {} detail".format(op_id))
        if re.search(failed_oper, output):
            log_install_errors(ctx, output)
            ctx.error("Operation {} failed".format(op_id))
            return  # for same of clarity

        ctx.info("Operation {} finished successfully".format(op_id))
        if has_tar is True:
            ctx.set_operation_id(ctx.software_packages, op_id)
            ctx.info("The operation {} stored".format(op_id))

        return  # for sake of clarity
    else:
        log_install_errors(ctx, output)
        ctx.error("Operation {} failed".format(op_id))


def install_activate_deactivate(ctx, cmd):
    message = "Waiting the operation to continue asynchronously"
    ctx.info(message)
    ctx.post_status(message)

    op_success = "The install operation will continue asynchronously"
    output = ctx.send(cmd, timeout=7200)
    result = re.search('Install operation (\d+) \'', output)
    if result:
        op_id = result.group(1)
    else:
        log_install_errors(ctx, output)
        ctx.error("Operation failed")
        return

    if op_success in output:
        success = watch_install(ctx, cmd, op_id)
        if not success:
            ctx.error("Reload or boot failure")
            return

        ctx.info("Operation {} finished successfully".format(op_id))
        return
    else:
        log_install_errors(ctx, output)
        ctx.error("Operation {} failed".format(op_id))
        return


def install_remove_all(ctx, cmd, hostname):
    """
    Success Condition:
    RP/0/RSP0/CPU0:RO#admin install remove inactive async

    Install operation 36 '(admin) install remove inactive' started by user 'root' via CLI at 01:14:44 PST Wed Feb 25 1987.
    / 1% complete: The operation can no longer be aborted (ctrl-c for options)
    Info:     This operation will remove the following package:
    Info:         disk0:asr9k-px-5.3.3.CSCuz14049-1.0.0
    - 1% complete: The operation can no longer be aborted (ctrl-c for options)
    \ 1% complete: The operation can no longer be aborted (ctrl-c for options)
    Info:     After this install remove the following install rollback point will no longer be reachable, as the required packages will not be present:
    Info:         15
    | 1% complete: The operation can no longer be aborted (ctrl-c for options)
    Proceed with removing these packages? [confirm]
    / 1% complete: The operation can no longer be aborted (ctrl-c for options)
    The install operation will continue asynchronously.
    RP/0/RSP0/CPU0:RO#Install operation 36 completed successfully at 01:14:51 PST Wed Feb 25 1987.

    Failed Conditions:
    RP/0/RSP0/CPU0:RO(admin)#install remove inactive
    Install operation 588 '(admin) install remove inactive' started by user 'lab'
    via CLI at 00:42:48 PST Tue Feb 24 1987.
    Error:    Cannot proceed with the remove operation because there are no
    Error:    packages that can be removed. Packages can only be removed if they
    Error:    are not part of the active software and not part of the committed
    Error:    software.
    Error:    Suggested steps to resolve this:
    Error:     - check the set of active packages using '(admin) show install
    Error:       active'.
    Error:     - check the committed software using '(admin) show install
    Error:       committed'.
    Error:     - check the set of inactive packages using '(admin) show install
    Error:       inactive'.
    Install operation 588 failed at 00:42:48 PST Tue Feb 24 1987.

    RP/0/RSP0/CPU0:RO(admin)#install remove inactive async
    Error:    Cannot proceed with the operation as another install command is
    Error:    currently in operation.
    Error:    Suggested steps to resolve this:
    Error:     - use 'show install request' to see the state of the current install
    Error:       operation.
    Error:     - re-issue the command when the current operation has completed.
    """

    global plugin_ctx
    plugin_ctx = ctx

    # no op_id is returned from XR for install remove inactive
    # need to figure out the last op_id first

    cmd_show_install_log_reverse = \
        'admin show install log reverse | utility egrep "Install operation [0-9]+ started"'
    output = ctx.send(cmd_show_install_log_reverse, timeout=300)

    if 'No log information' in output:
        op_id = 0
    else:
        result = re.search('Install operation (\d+) started', output)
        if result:
            op_id = int(result.group(1))
        else:
            log_install_errors(ctx, output)
            ctx.error("Operation ID not found by admin show install log reverse")
            return

    # Expected Operation ID
    op_id += 1

    operr = "Install operation {} failed at".format(op_id)
    Error1 = re.compile("Error:     - re-issue the command when the current operation has completed.")
    Error2 = re.compile(operr)
    Proceed_removing = re.compile("\[confirm\]")
    Host_prompt = re.compile(hostname)

    events = [Host_prompt, Error1, Error2, Proceed_removing]
    transitions = [
        (Error1, [0], -1, CommandError("Another install command is currently in operation", hostname), 1800),
        (Error2, [0], -1, CommandError("No packages can be removed", hostname), 1800),
        (Proceed_removing, [0], 2, send_yes, 1800),
        (Host_prompt, [2], -1, None, 1800),
    ]

    if not ctx.run_fsm("Remove Inactive All", cmd, events, transitions, timeout=1800):
        ctx.error("Failed: {}".format(cmd))

    message = "Waiting the operation to continue asynchronously"
    ctx.info(message)
    ctx.post_status(message)

    last_status = None
    no_install = r"There are no install requests in operation"
    op_progress = r"The operation is (\d+)% complete"
    cmd_show_install_request = "admin show install request"
    op_success = "Install operation {} completed successfully".format(op_id)
    propeller = itertools.cycle(["|", "/", "-", "\\", "|", "/", "-", "\\"])

    finish = False
    time_tried = 0
    op_id = str(op_id)
    while not finish:
        try:
            try:
                # this is to catch the successful operation as soon as possible
                ctx.send("", wait_for_string=op_success, timeout=20)
                finish = True
            except ctx.CommandTimeoutError:
                pass

            message = ""
            # on CRS, it is observed that during Add, any command typed hangs for a while
            output = ctx.send(cmd_show_install_request, timeout=300)
            if op_id in output:
                result = re.search(op_progress, output)
                if result:
                    status = result.group(0)
                    message = "{} {}".format(propeller.next(), status)

                if message != last_status:
                    ctx.post_status(message)
                    last_status = message
        except (ConnectionError, ctx.CommandTimeoutError) as e:
            if time_tried > 120:
                raise e

            time_tried += 1
            time.sleep(30)

        if no_install in output:
            break

    cmd_show_install_log = "admin show install log {} detail".format(op_id)
    output = ctx.send(cmd_show_install_log, timeout=300)
    ctx.info(output)

    if op_success in output:
        message = "Remove All Inactive Package(s) Successfully"
        ctx.info(message)
        ctx.post_status(message)
    else:
        ctx.error("Remove All Inactive Package(s) failed")



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

"""
NCS4K

Production Packages

External Names                                Internal Names
ncs4k-full-x.iso-6.0.2
ncs4k-mini-x.iso-6.0.2
ncs4k-k9sec.pkg-6.0.2
ncs4k-mpls.pkg-6.0.2
ncs4k-mcast.pkg-6.0.2
ncs4k-mgbl.pkg-6.0.2

NCS6K

Production Packages

External Names                                Internal Names
ncs6k-doc.pkg-5.2.4                           ncs6k-doc-5.2.4
ncs6k-li.pkg-5.2.4                            ncs6k-li-5.2.4
ncs6k-mcast.pkg-5.2.4                         ncs6k-mcast-5.2.4
ncs6k-mgbl.pkg-5.2.4                          ncs6k-mgbl-5.2.4
ncs6k-mini-x.iso-5.2.4                        ncs6k-mini-x-5.2.4
ncs6k-mpls.pkg-5.2.4                          ncs6k-mpls-5.2.4
ncs6k-sysadmin.iso-5.2.4                      ncs6k-sysadmin-5.2.4
ncs6k-full-x.iso-5.2.4                        ncs6k-full-x-5.2.4
ncs6k-5.2.5.CSCuy47880.smu                    ncs6k-5.2.5.CSCuy47880-1.0.0  <- subversion added

Engineering Packages

External Names                                Internal Names
ncs6k-mcast.pkg-5.2.5.47I.DT_IMAGE            ncs6k-mcast-5.2.5.47I
ncs6k-mini-x.iso-6.1.0.07I.DT_IMAGE           ncs6k-xr-5.2.5.47I
ncs6k-5.2.5.47I.CSCuy47880-0.0.4.i.smu        ncs6k-5.2.5.47I.CSCuy47880-0.0.4.i

ASR9K-64

Production Packages - not finalized yet

External Names                                Internal Names
asr9k-mcast-x64-2.0.0.0-r611.x86_64.rpm       asr9k-mcast-x64-2.0.0.0-r611
asr9k-bgp-x64-1.0.0.0-r611.x86_64.rpm         asr9k-bgp-x64-1.0.0.0-r611
asr9k-mgbl-x64-3.0.0.0-r611.x86_64.rpm        asr9k-mgbl-x64-3.0.0.0-r611
asr9k-full-x64.iso-6.1.1                      asr9k-xr-6.1.1
asr9k-mini-x64.iso-6.1.1                      asr9k-xr-6.1.1

Engineering Packages

External Names                                                          Internal Names
asr9k-mcast-x64-2.0.0.0-r61116I.x86_64.rpm-6.1.1.16I.DT_IMAGE           asr9k-mcast-x64-2.0.0.0-r61116I
asr9k-bgp-x64-1.0.0.0-r61116I.x86_64.rpm-6.1.1.16I.DT_IMAGE             asr9k-bgp-x64-1.0.0.0-r61116I
asr9k-mgbl-x64-3.0.0.0-r61116I.x86_64.rpm-6.1.1.16I.DT_IMAGE            asr9k-mgbl-x64-3.0.0.0-r61116I
asr9k-full-x64.iso-6.1.1.16I.DT_IMAGE                                   asr9k-full-x64-6.1.1.16I
asr9k-mini-x64.iso-6.1.1.16I.DT_IMAGE                                   asr9k-mini-x64-6.1.1.16I

NCS5K

Production Packages

External Names                                Internal Names
ncs5k-sysadmin.iso-6.0.1                      ncs5k-sysadmin-6.0.1
ncs5k-full-x.iso-6.0.1                        ncs5k-xr-6.0.1
ncs5k-mini-x.iso-6.0.1                        ncs5k-xr-6.0.1
ncs5k-mcast-2.0.0.0-r601.x86_64.rpm-6.0.1     ncs5k-mcast-2.0.0.0-r601
ncs5k-mgbl-2.0.0.0-r601.x86_64.rpm-6.0.1      ncs5k-mgbl-2.0.0.0-r601
ncs5k-mpls-2.0.0.0-r601.x86_64.rpm-6.0.1      ncs5k-mpls-2.0.0.0-r601
ncs5k-k9sec-2.0.0.0-r601.x86_64.rpm-6.0.1     ncs5k-k9sec-2.0.0.0-r601
ncs5k-isis-2.0.0.0-r601.x86_64.rpm-6.0.1      ncs5k-isis-2.0.0.0-r601
ncs5k-ospf-2.0.0.0-r601.x86_64.rpm-6.0.1      ncs5k-ospf-2.0.0.0-r601

Engineering Packages

External Names                                                Internal Names
ncs5k-mgbl-x64-3.0.0.0-r61116I.x86_64.rpm-6.0.1.16I.DT_IMAGE  ncs5k-mgbl-3.0.0.0-r60116I
ncs5k-sysadmin.iso-6.0.1                                      ncs5k-sysadmin-6.0.1.26I
ncs5k-full-x.iso-6.0.1.16I.DT_IMAGE                           ncs5k-xr-6.0.1.16I

NCS5500

Production Packages

External Names                                                Internal Names
ncs5500-eigrp-2.0.0.0-r601.x86_64.rpm-6.0.1                   ncs5500-eigrp-2.0.0.0-r601
ncs5500-isis-2.0.0.0-r601.x86_64.rpm-6.0.1                    ncs5500-isis-2.0.0.0-r601
ncs5500-k9sec-2.0.0.0-r601.x86_64.rpm-6.0.1                   ncs5500-k9sec-2.0.0.0-r601
ncs5500-m2m-2.0.0.0-r601.x86_64.rpm-6.0.1                     ncs5500-m2m-2.0.0.0-r601
ncs5500-mgbl-3.0.0.0-r601.x86_64.rpm-6.0.1                    ncs5500-mgbl-3.0.0.0-r601
ncs5500-mini-x.iso-6.0.1                                      ncs5500-xr-6.0.1
ncs5500-mpls-te-rsvp-2.0.0.0-r601.x86_64.rpm-6.0.1            ncs5500-mpls-te-rsvp-2.0.0.0-r601
ncs5500-mpls-2.0.0.0-r601.x86_64.rpm-6.0.1                    ncs5500-mpls-2.0.0.0-r601
ncs5500-ospf-1.0.0.0-r601.x86_64.rpm-6.0.1                    ncs5500-ospf-1.0.0.0-r601
ncs5500-parser-1.0.0.0-r601.x86_64.rpm-6.0.1                  ncs5500-parser-1.0.0.0-r601
"""
import re

platforms = ['asr9k', 'ncs1k', 'ncs4k', 'ncs5k', 'ncs5500', 'ncs6k', 'xrv9k']


version_dict = {"asr9k ncs1k ncs5k ncs5500 xrv9k":  # r61117I or r611 or 6.1.1.17I or 6.1.1
                re.compile("(?P<VERSION>(r\d+\d+\d+(\d+\w+)?)|(\d+\.\d+\.\d+(\.\d+\w+)?)(?!\.\d)(?!-))"),
                "ncs4k ncs6k":                      # 5.2.4 or 5.2.4.47I
                re.compile("(?P<VERSION>\d+\.\d+\.\d+(\.\d+\w+)?)"),
                }

smu_re = re.compile("(?P<SMU>CSC[a-z]{2}\d{5})")

subversion_dict = {"asr9k ncs1k ncs5k ncs5500 xrv9k":
                   re.compile("-(?P<SUBVERSION>\d+\.\d+\.\d+\.\d+)-"),  # 2.0.0.0
                   "ncs4k ncs6k":
                   re.compile("CSC.*(?P<SUBVERSION>\d+\.\d+\.\d+?)"),   # 0.0.4
                   }


class SoftwarePackage(object):
    def __init__(self, package_name):
        self.package_name = package_name

        self._platform = None
        self._package_type = None
        self._version = None
        self._smu = None
        self._subversion = None

    @property
    def platform(self):
        if not self._platform:
            for platform in platforms:
                if platform + "-" in self.package_name:
                    self._platform = platform
                    break

        return self._platform

    @property
    def package_type(self):
        if not self._package_type:
            # For ASR9K-X64, NCS1K, NCS5K, NCS5500:
            #     Extract the package type string before X.X.X.X
            # For NCS6K
            #     Extract the package type string before X.X.X
            pattern = '-\d+\.\d+\.\d+' if self.platform == 'ncs6k' or \
                self.platform == 'ncs4k' else '-\d\.\d\.\d.\d'

            if self.platform and self.platform in self.package_name:
                match = re.search(pattern, self.package_name)

                # Special handling for mini, full, and sysadmin ISO on ASR9K-X64, NCS1K, NCS5K, NCS5500
                # Example: ncs5500-mini-x.iso-6.0.1, asr9k-full-x64.iso-6.1.1
                # Package type string is before the 3 part version string

                # External Name: ncs5k-goldenk9-x.iso-6.3.1.11I.0, Internal Name: ncs5k-goldenk9-x-6.3.1.11I
                if not match and sum([x in self.package_name for x in ['full', 'mini', 'sysadmin', 'goldenk9']]) > 0:
                    # Use the three part match for these ISO packages
                    match = re.search('-\d+\.\d+\.\d+', self.package_name)

                if match:
                    # Extract the package type
                    self._package_type = self.package_name[0:match.start()].replace(self.platform + '-', '')

                if self._package_type:
                    # Takes care the external to internal name matching
                    # Example, ncs6k-mgbl.pkg-5.2.5 -> mgbl, ncs5500-mini-x.iso-6.0.1 -> mini-x
                    self._package_type = self._package_type.replace('.pkg', '').replace('.iso', '')

        return self._package_type

    @property
    def version(self):
        if not self._version:
            dict_values = self.get_values(version_dict, self.platform)
            if self.platform and dict_values:
                to_match = self.package_name.replace(self.platform, '')
                result = re.search(dict_values, to_match)
                if result:
                    self._version = result.group("VERSION")

        return self._version

    @property
    def smu(self):
        if not self._smu:
            result = re.search(smu_re, self.package_name)
            if result:
                self._smu = result.group("SMU")

        return self._smu

    @property
    def subversion(self):
        if not self._subversion:
            dict_values = self.get_values(subversion_dict, self.platform)
            if self.platform and dict_values:
                # For NCS6K, only need to consider subversion if it is a SMU.
                if self.platform in ["asr9k", "ncs1k", "ncs5k", "ncs5500", "xrv9k"] or self.smu:
                    to_match = self.package_name.replace(self.platform, '')
                    result = re.search(dict_values, to_match)
                    if result:
                        self._subversion = result.group("SUBVERSION")

        return self._subversion

    def get_values(self, dictionary, key):
        for keys in dictionary.keys():
            if key in keys.split():
                return dictionary.get(keys)
        return None

    def is_valid(self):
        return self.platform and self.version and (self.package_type or self.smu)

    def __eq__(self, other):
        result = self.platform == other.platform and \
            (self.package_type == other.package_type) and \
            self.version == other.version and \
            self.smu == other.smu and \
            (self.subversion == other.subversion if self.subversion and other.subversion else True)

        return result

    def __hash__(self):
        return hash("{}{}{}{}{}".format(
            self.platform, self.package_type, self.version, self.smu, self.subversion))

    @staticmethod
    def from_show_cmd(cmd):
        software_packages = set()
        data = cmd.split()

        for line in data:
            software_package = SoftwarePackage(line)
            if software_package.is_valid():
                software_packages.add(software_package)
        return software_packages

    @staticmethod
    def from_package_list(pkg_list):
        software_packages = set()
        for pkg in pkg_list:
            software_package = SoftwarePackage(pkg)
            if software_package.is_valid():
                """ for debugging
                print('package_name', software_package.package_name,
                      'platform', software_package.platform, 'package_type', software_package.package_type,
                      'version', software_package.version, 'smu', software_package.smu,
                      'subversion', software_package.subversion)
                """
                software_packages.add(software_package)
        return software_packages

    def __repr__(self):
        return self.package_name

    def __str__(self):
        return self.__repr__()



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

from csmpe.plugins import CSMPlugin
from install import observe_install_add_remove
from install import check_ncs6k_release, check_ncs4k_release
from csmpe.core_plugins.csm_get_inventory.exr.plugin import get_package, get_inventory

import re


class Plugin(CSMPlugin):
    """This plugin adds packages from repository to the device."""
    name = "Install Add Plugin"
    platforms = {'ASR9K', 'NCS1K', 'NCS4K', 'NCS5K', 'NCS5500', 'NCS6K', 'IOSXRv-9K', 'IOSXRv-X64'}
    phases = {'Add'}
    os = {'eXR'}

    def install_add(self, server_repository_url, s_packages, has_tar=False):
        """
        Success Condition:
        ADD for tftp/local:
        install add source tftp://223.255.254.254/auto/tftpboot-users/alextang/ ncs6k-mpls.pkg-6.1.0.07I.DT_IMAGE
        May 24 18:54:12 Install operation will continue in the background
        RP/0/RP0/CPU0:Deploy#May 24 18:54:30 Install operation 12 finished successfully

        ADD for sftp/ftp
        RP/0/RP0/CPU0:Deploy-2#install add source sftp://terastream@172.20.168.195/echami ncs6k-li.pkg-5.2.5-V2

        Jun 20 18:58:08 Install operation 38 started by root:
         install add source sftp://terastream:cisco@172.20.168.195/echami ncs6k-li.pkg-5.2.5-V2
        password:Jun 20 18:58:24 Install operation will continue in the background
        RP/0/RP0/CPU0:Deploy-2#
        """
        if server_repository_url.startswith("sftp://") or server_repository_url.startswith("ftp://"):

            rest_of_url = server_repository_url[server_repository_url.index('//') + 2:]
            password = re.search(r':(.*)@', rest_of_url).group(1)
            url_without_password = server_repository_url.replace(':' + password, '')

            cmd = "install add source {} {}".format(url_without_password, s_packages)

            output1 = self.ctx.send(cmd, wait_for_string="[Pp]assword:", timeout=60)
            output2 = self.ctx.send(password, timeout=100, password=True)
            output = output1 + output2

        elif server_repository_url.startswith("scp"):
            # scp username:password@x.x.x.x:/home_directory destination_on_host
            scp_username_and_password, sep, server_and_directory_and_destination = server_repository_url.partition('@')
            # scp_username_and_password = 'scp username:password', sep = '@',
            # server_ip_and_directory = 'x.x.x.x:/home_directory destination_on_host'
            if not scp_username_and_password or not sep or not server_and_directory_and_destination:
                self.ctx.error("Check if the SCP server repository is configured correctly on CSM Server.")

            scp_username, sep, scp_password = scp_username_and_password.partition(':')
            if not scp_username or not sep or not scp_password:
                self.ctx.error("Check if the SCP server repository is configured correctly on CSM Server.")

            server_and_directory, sep, destination_on_host = server_and_directory_and_destination.partition(' ')
            if not server_and_directory or not sep or not destination_on_host:
                self.ctx.error("Check if the SCP server repository is configured correctly on CSM Server.")

            # scp username:@x.x.x.x:/home_directory
            url = scp_username + '@' + server_and_directory
            for package in s_packages.split():
                cmd = "{}/{} {}".format(url, package, destination_on_host)
                output1 = self.ctx.send(cmd, wait_for_string="[Pp]assword:", timeout=60)
                output2 = self.ctx.send(scp_password, timeout=100)

            cmd = "install add source {} {}".format(destination_on_host, s_packages)
            output = self.ctx.send(cmd, timeout=100)
        else:
            cmd = "install add source {} {}".format(server_repository_url, s_packages)
            output = self.ctx.send(cmd, timeout=100)

        observe_install_add_remove(self.ctx, output, has_tar=has_tar)

    def run(self):
        check_ncs6k_release(self.ctx)
        check_ncs4k_release(self.ctx)

        server_repository_url = self.ctx.server_repository_url
        if server_repository_url is None:
            self.ctx.error("No repository provided")
            return

        packages = self.ctx.software_packages
        if packages is None:
            self.ctx.error("No package list provided")
            return

        has_tar = False

        if self.ctx.family == 'NCS6K' or self.ctx.family == 'NCS4K':
            s_packages = " ".join([package for package in packages
                                   if ('iso' in package or 'pkg' in package or 'smu' in package or 'tar' in package)])
        else:
            s_packages = " ".join([package for package in packages
                                   if ('rpm' in package or 'iso' in package or 'tar' in package)])

        if 'tar' in s_packages:
            has_tar = True

        if not s_packages:
            self.ctx.error("None of the selected package(s) has an acceptable file extension.")

        self.ctx.info("Add Package(s) Pending")
        self.ctx.post_status("Add Package(s) Pending")

        self.install_add(server_repository_url, s_packages, has_tar=has_tar)

        self.ctx.info("Package(s) Added Successfully")

        # Refresh package and inventory information
        get_package(self.ctx)
        get_inventory(self.ctx)



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

from csmpe.plugins import CSMPlugin
from csmpe.core_plugins.csm_get_inventory.ios.plugin import get_package, get_inventory
from install import remove_exist_image


class Plugin(CSMPlugin):
    """This plugin removes inactive packages from the device."""
    name = "Install Remove Plugin"
    platforms = {'ASR900'}
    phases = {'Remove'}
    os = {'IOS'}

    def run(self):
        packages_to_remove = self.ctx.software_packages
        if packages_to_remove is None:
            self.ctx.error("No package list provided")
            return

        self.ctx.info("Remove Package(s) Pending")
        self.ctx.post_status("Remove Package(s) Pending")

        for pkg in packages_to_remove:
            self.ctx.info("Delete package flash:{}".format(pkg))

            package = 'flash:' + pkg
            output = remove_exist_image(self.ctx, package)

            if not output:
                self.ctx.info("flash:{} Removal Failed".format(pkg))
                break
        else:
            self.ctx.info("Package(s) Removed Successfully")
            # Refresh package and inventory information
            get_package(self.ctx)
            get_inventory(self.ctx)
            return

        self.ctx.error("Package(s) Removal Failed")
        return



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================
import re
from time import time
from datetime import datetime
from csmpe.plugins import CSMPlugin
from condoor.exceptions import CommandSyntaxError


class Plugin(CSMPlugin):
    """This plugin checks the ISIS neighbor."""
    name = "ISIS Neighbor Check Plugin"
    platforms = {'ASR9K', 'CRS', 'NCS1K', 'NCS4K', 'NCS5K', 'NCS5500', 'NCS6K', 'IOSXRv-9K', 'IOSXRv-X64'}
    phases = {'Pre-Upgrade', 'Post-Upgrade'}

    def run(self):
        """
        RP/0/RP0/CPU0:#show isis neighbor summary
        Thu May 19 18:06:11.239 UTC

        IS-IS isp neighbor summary:
        State         L1       L2     L1L2
        Up             0        0        1
        Init           0        0        0
        Failed         0        0        0

        This plugin check the number of ISIS Neighbors and store this information in format
        {
            <instance>: {
                "Up": [ <L1>, <L2>, <L1/L2> ],
                "Init": [ <L1>, <L2>, <L1/L2> ],
                "Failed": [ <L1>, <L2>, <L1/L2> ]
            }
        }
        """
        cmd = "show isis neighbor summary"
        isis_neighbor_info = {}
        try:
            output = self.ctx.send(cmd)
        except CommandSyntaxError:
            # This will happen when the device is ASR9K-X64 and the ISIS package is not installed.
            self.ctx.info("The CLI 'show isis neighbor summary' is not available for checking the number " +
                          "of ISIS neighbors. Possible reason: the ISIS package (if any) needs to be installed.")
            return

        if output:
            isis_instance = None
            for line in output.split('\n'):
                result = re.search('IS-IS (.*) neighbor summary:', line)
                if result:
                    isis_instance = result.group(1)
                    isis_neighbor_info[isis_instance] = {}
                    continue
                result = re.search('Up\s+(\d+)\s+(\d+)\s+(\d+)', line)
                if result and isis_instance:
                    isis_neighbor_info[isis_instance]["Up"] = [result.group(n) for n in range(1, 4)]
                    continue
                result = re.search('Init\s+(\d+)\s+(\d+)\s+(\d+)', line)
                if result and isis_instance:
                    isis_neighbor_info[isis_instance]["Init"] = [result.group(n) for n in range(1, 4)]
                    continue
                result = re.search('Failed\s+(\d+)\s+(\d+)\s+(\d+)', line)
                if result and isis_instance:
                    isis_neighbor_info[isis_instance]["Failed"] = [result.group(n) for n in range(1, 4)]
                    continue

            if isis_instance:
                self.ctx.info("There is {} ISIS protocol instance(s) active".format(len(isis_neighbor_info)))
                for instance, state_dict in isis_neighbor_info.items():
                    for state, neighbors in state_dict.items():
                        self.ctx.info("Instance {} {:<6} L1={} L2={} L1L2={}".format(instance, state, *neighbors))

                filename = self.ctx.save_to_file(cmd, output)
                if filename:
                    self.ctx.info("The '{}' command output saved to {}".format(cmd, filename))

                if self.ctx.phase == "Pre-Upgrade":
                    self.ctx.save_data("isis_neighbors", isis_neighbor_info)
                    if filename:
                        # store the full_path to command output under the cmd key
                        self.ctx.save_data(cmd, filename)

            else:
                self.ctx.info("No ISIS protocol instance active")

            if self.ctx.phase == "Post-Upgrade":
                self.compare_data("isis_neighbors", isis_neighbor_info)

    def compare_data(self, storage_key, current_data):
        levels = ["L1", "L2", "L1L2"]
        previous_data, timestamp = self.ctx.load_data(storage_key)
        if previous_data is None:
            self.ctx.warning("No data stored from Pre-Upgrade phase. Can't compare.")
            return

        if previous_data:
            self.ctx.info("Pre-Upgrade phase data available for comparison")
            self.ctx.info("Pre-Upgrade data collected on {}".format(
                datetime.fromtimestamp(int(timestamp)).strftime('%Y-%m-%d %H:%M:%S')))
            if timestamp < time() - (60 * 60 * 2):  # two hours
                self.ctx.warning("Pre-Upgrade phase data older than 2 hours")

            for previous_instance, previous_state_dict in previous_data.items():
                try:
                    current_state_dict = current_data[previous_instance]
                except KeyError:
                    self.ctx.warning("No ISIS instance '{}' detected after upgrade".format(previous_instance))
                    continue

                for state, previous_neighbors in previous_state_dict.items():
                    current_neighbors = current_state_dict[state]
                    if previous_neighbors != current_neighbors:
                        for level in range(3):
                            if previous_neighbors[level] != current_neighbors[level]:
                                self.ctx.warning(
                                    "Number of ISIS neighbors for instance '{}' in state {} at {} is different"
                                    "Pre-Install={} Post-Install={}".format(
                                        previous_instance, state, levels[level], previous_neighbors[level],
                                        current_neighbors[level]
                                    )
                                )
                else:
                    self.ctx.info("Number of ISIS neighbors for instance '{}' is the "
                                  "same as during pre-install check".format(previous_instance))
        else:
            self.ctx.warning("No data stored from Pre-Upgrade phase")
            return



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================


import os

from csmpe.plugins import CSMPlugin
from utils import get_filesystems


class Plugin(CSMPlugin):
    """This plugin captures custom commands and stores in the log directory."""
    name = "Disk Space Check Plugin"
    platforms = {'ASR9K', 'CRS'}
    phases = {'Pre-Add'}
    os = {'XR'}

    def _get_pie_size(self, package_url):
        """
        # admin show install pie-info ftp://terastream:*****@172.20.168.195//home/terastream/4.3.1FCS/asr9k-doc-px.pie-4.3.1
        Contents of pie file '/ftp://terastream:*****@172.20.168.195//home/terastream/4.3.1FCS/asr9k-doc-px.pie-4.3.1':
            Expiry date       : < NOT AVAILABLE
            Uncompressed size : 25792000
            Compressed size   : 5602232
            Size on disk      : 5668352

            asr9k-doc-px-4.3.1
                asr9K-doc-supp-4.3.1
        """
        cmd = "admin show install pie-info " + package_url
        output = self.ctx.send(cmd, timeout=3600)
        if output:
            for line in output.split('\n'):
                if "Uncompressed" in line:
                    size = long(line.split(":")[1].strip())
                    return size
                if line and line[:6] == "Error:":
                    self.ctx.error(output)

    def run(self):
        try:
            packages = self.ctx.software_packages
        except AttributeError:
            self.ctx.warning("No package list provided. Skipping calculation of required free harddisk space.")
            return

        try:
            server_repository_url = self.ctx.server_repository_url
        except AttributeError:
            self.ctx.warning("No repository path provided.")
            return

        if server_repository_url[:4] != 'tftp' and server_repository_url[:3] != 'ftp':
            self.ctx.info('Skipping as disk space check only works for TFTP/FTP.')
            return

        file_systems = get_filesystems(self.ctx)
        disk0 = file_systems.get('disk0:', None)
        if disk0:
            disk0_free = disk0.get('free', 0)
        else:
            self.ctx.error("No filesystem 'disk0:' on active RP.")
            return

        total_size = 0
        for package in packages:
            if package == "":
                continue

            if 'pie' not in package:
                self.ctx.info("Package: {} cannot be checked, disk space check result will not be accurate.".format(package))
                continue

            package_url = os.path.join(server_repository_url, package)
            size = self._get_pie_size(package_url)
            total_size += size
            self.ctx.info("Package: {} requires {} bytes.".format(package, size))

        if disk0_free < total_size:
            self.ctx.error("Not enough space on disk0: to install packages. The install process can't proceed."
                           "\n"
                           "Total (required/available): {}/{} bytes".format(total_size, disk0_free))
        else:
            self.ctx.info("There is enough space on disk0: to install packages.")
            self.ctx.info("Total (required/available): {}/{} bytes".format(total_size, disk0_free))
        return True



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import re

# from documentation:
# http://www.cisco.com/c/en/us/td/docs/routers/asr9000/software/asr9k_r5-3/sysman/configuration/guide/b-sysman-cg-53xasr9k/b-sysman-cg-53xasr9k_chapter_0100.html#con_57141

"""
For example,
disk0:asr9k-px-5.3.3.CSCuz33376-1.0.0

package_type = None
version_re = 5.3.3
smu_re = CSCuz33376
sp_re = None
subversion_re = 1.0.0
"""
platforms = ["asr9k", "hfr"]
# 'services-infra' needs to be before 'service' for matching purpose
package_types = "mini mcast mgbl mpls k9sec diags fpd doc bng li optic services-infra services " \
                "infra-test video 9000v asr901 asr903 ncs500x".split()
version_re = re.compile("(?P<VERSION>\d+\.\d+\.\d+(\.\d+\w+)?)")
smu_re = re.compile("(?P<SMU>CSC[a-z]{2}\d{5})")
sp_re = re.compile("(?P<SP>(sp|fp)\d{0,2})")
subversion_re = re.compile("(CSC|sp|fp).*(?P<SUBVERSION>\d+\.\d+\.\d+?)")


class SoftwarePackage(object):
    def __init__(self, package_name):
        # Special logic to handle these two packages.
        # External Names:                           Internal Names:
        #     asr9k-asr9000v-nV-px.pie-6.1.2        asr9k-9000v-nV-px-6.1.2
        #     asr9k-services-infra-px.pie-6.1.2     asr9k-services-infra-6.1.2
        if 'asr9000v' in package_name:
            package_name = package_name.replace('asr9000v', '9000v')
        elif 'services-infra-px' in package_name:
            package_name = package_name.replace('-px', '')

        self.package_name = package_name

    @property
    def platform(self):
        for platform in platforms:
            if platform + "-" in self.package_name:
                return platform
        else:
            return None

    @property
    def package_type(self):
        for package_type in package_types:
            if "-" + package_type in self.package_name:
                return package_type
        else:
            return None

    @property
    def version(self):
        result = re.search(version_re, self.package_name)
        return result.group("VERSION") if result else None

    @property
    def smu(self):
        result = re.search(smu_re, self.package_name)
        return result.group("SMU") if result else None

    @property
    def sp(self):
        result = re.search(sp_re, self.package_name)
        return result.group("SP") if result else None

    @property
    def subversion(self):
        if self.sp or self.smu:
            result = re.search(subversion_re, self.package_name)
            return result.group("SUBVERSION") if result else None
        return None

    def is_valid(self):
        return self.platform and self.version and (self.package_type or self.smu or self.sp)

    def __eq__(self, other):
        result = self.platform == other.platform and \
            self.package_type == other.package_type and \
            self.version == other.version and \
            self.smu == other.smu and \
            self.sp == other.sp and \
            (self.subversion == other.subversion if self.subversion and other.subversion else True)

        if result:
            # Append the disk location to the package name
            if ":" in self.package_name:
                disk = self.package_name.split(':')[0] + ":"
                if not other.package_name.startswith(disk):
                    other.package_name = disk + other.package_name

        return result

    def __hash__(self):
        return hash("{}{}{}{}{}".format(
            self.platform, self.package_type, self.version, self.smu, self.sp, self.subversion))

    @staticmethod
    def from_show_cmd(cmd):
        software_packages = set()
        data = cmd.split()
        for line in data:
            software_package = SoftwarePackage(line)
            if software_package.is_valid():
                software_packages.add(software_package)
        return software_packages

    @staticmethod
    def from_package_list(pkg_list):
        software_packages = set()
        for pkg in pkg_list:
            software_package = SoftwarePackage(pkg)
            if software_package.is_valid():
                software_packages.add(software_package)

        return software_packages

    def __repr__(self):
        return self.package_name

    def __str__(self):
        return self.__repr__()



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================


# from package_lib import SoftwarePackage
import re
from time import time
from datetime import datetime
from csmpe.plugins import CSMPlugin
from install import expand_subpkgs
from install import install_activate_reload
from install import install_activate_write_memory
from install import install_activate_issu
from csmpe.core_plugins.csm_get_inventory.ios_xe.plugin import get_package, get_inventory
from csmpe.core_plugins.csm_install_operations.utils import update_device_info_udi
from utils import remove_exist_image
from utils import xe_show_platform
from utils import install_add_remove
from utils import installed_package_name


class Plugin(CSMPlugin):
    """This plugin Activates packages on the device."""
    name = "Install Activate Plugin"
    platforms = {'ASR900'}
    phases = {'Activate'}
    os = {'XE'}

    def run(self):
        """
        Performs install activate operation
        """

        rsp_count = int(self.ctx.load_data('xe_rsp_count')[0])
        pkg = self.ctx.load_data('xe_activate_pkg')[0]
        mode = self.ctx.load_data('xe_boot_mode')[0]
        folder = self.ctx.load_data('xe_install_folder')[0]

        self.ctx.info("Activate number of RSP = {}".format(rsp_count))
        self.ctx.info("Activate package = {}".format(pkg))
        self.ctx.info("Activate package mode = {}".format(mode))
        self.ctx.info("Install folder = {}".format(folder))

        self.ctx.info("Activate package(s) pending")
        self.ctx.post_status("Activate Package(s) Pending")

        # prompt = self.ctx._connection.hostname

        # issu: need to copy the consolidated image to the installed folder
        if mode == 'issu':
            cmd = 'copy bootflash:' + pkg + ' ' + folder + '/' + pkg
            install_add_remove(self.ctx, cmd)
            # package = 'bootflash:' + pkg
            # remove_exist_image(self.ctx, package)

        # subpackage: need to expand the consolidated image to the installed folder
        if mode == 'subpackage':
            result = expand_subpkgs(self.ctx, rsp_count, folder, pkg)
            if not result:
                self.ctx.error("Error in extracting sub-images from the consolidated "
                               "image {}".format(pkg))
                return

        # configurations
        cmd = "configure terminal"
        self.ctx.send(cmd)
        cmd = "config-register 0x2102"
        self.ctx.send(cmd)

        if mode == 'issu':
            cmd = 'redundancy'
            self.ctx.send(cmd)
            cmd = 'mode sso'
            self.ctx.send(cmd)
        else:
            cmd = "no boot system"
            self.ctx.send(cmd)
            if mode == 'consolidated':
                cmd = "boot system bootflash:" + pkg
            else:
                cmd = 'boot system ' + folder + '/packages.conf'
            self.ctx.send(cmd)

        self.ctx.send('end')

        cmd = "write memory"
        install_activate_write_memory(self.ctx, cmd)
        # self.ctx.send(cmd, timeout=300, wait_for_string=prompt)

        # Start activation
        if mode == 'issu':
            cmd = 'request platform software package install node file ' + \
                  folder + '/' + pkg + ' interface-module-delay 160'
            install_activate_issu(self.ctx, cmd)
        else:
            install_activate_reload(self.ctx)

        self.ctx.info("Activate package done")

        # Refresh package and inventory information
        get_package(self.ctx)
        get_inventory(self.ctx)
        update_device_info_udi(self.ctx)

        # Verify the version
        activate_success = True
        if self.ctx._connection.os_version not in pkg:
            activate_success = False
            self.ctx.warning('The post-activate OS Version: {} while Activate package = '
                             '{}'.format(self.ctx._connection.os_version, pkg))

        # Verify the Image type
        if mode == 'issu' or mode == 'subpackage':
            pkg_conf = folder + '/packages.conf'
            image_type = installed_package_name(self.ctx, pkg_conf)
            if not image_type:
                activate_success = False
                self.ctx.warning('{} does not exist.'.format(pkg_conf))

            if image_type not in pkg:
                activate_success = False
                self.ctx.warning('The post-activate image type: {} while Activate package = '
                                 '{}'.format(image_type, pkg))
        else:
            # mode is consolidated
            output = self.ctx.send('show version | include ^System image')
            if output:
                m = re.search('"bootflash:(.*)"$', output)
                if m:
                    image_type = m.group(1)
                    if image_type not in pkg:
                        self.ctx.warning('The post-activate image type: {} while Activate '
                                         'package = {}'.format(image_type, pkg))
                else:
                    activate_success = False
                    self.ctx.info('show version = {}'.format(output))
                    self.ctx.warning('System image not found in show version: {}'.format(output))
            else:
                activate_success = False
                self.ctx.info('show version = {}'.format(output))
                self.ctx.warning('System image not found in show version: {}'.format(output))

        if not activate_success:
            # Refresh package information
            get_package(self.ctx)
            update_device_info_udi(self.ctx)
            self.ctx.warning('Activte image {} may have failed. Please see above warnings'.format(pkg))

        if mode == 'issu':
            # Remove all-in-one image from the installed folder
            package = folder + '/' + pkg
            remove_exist_image(self.ctx, package)
            if rsp_count == 2:
                package = 'stby-' + package
                remove_exist_image(self.ctx, package)

        # Verify the status of RP and SIP
        previous_data, timestamp = self.ctx.load_data('xe_show_platform')
        self.ctx.info("Pre-Activate data collected on {}".format(
            datetime.fromtimestamp(int(timestamp)).strftime('%Y-%m-%d %H:%M:%S')))
        if timestamp < time() - (60 * 60 * 2):  # two hours
            self.ctx.warning("Pre-Activate phase 'show platform' "
                             "data older than 2 hours")

        current_data = xe_show_platform(self.ctx)
        if not current_data:
            self.ctx.warning("The CLI 'show platform' is not able to determine the status of RP and SIP ")
            return

        for Slot, Status in previous_data.items():
            Type = Status[0]
            previous_state = Status[1]
            current_state = current_data[Slot][1]
            if previous_state != current_state:
                if previous_state == 'ok, active' and current_state == 'ok, standby' \
                        or previous_state == 'ok, standby' and current_state == 'ok, active':
                    continue
                self.ctx.warning("Slot {} Type {} state changes after Activation".format(Slot, Type))
                self.ctx.warning("\t Pre-Activate State = {} vs. Post-Activate State = "
                                 "{}".format(previous_state, current_state))
            if 'ok' not in current_state:
                self.ctx.warning("Slot {} Type {} is not in 'ok' state after "
                                 "activation".format(Slot, Type))

        self.ctx.info("The status of RP and SIP has been verified. Please check any warnings in plugins.log")



# =============================================================================
# plugin
#
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

from csmpe.plugins import CSMPlugin


class Plugin(CSMPlugin):
    """
    This plugin checks if the node redundancy is sufficient to proceed the upgrade.
    """
    name = "Node Redundancy Check Plugin"
    platforms = {'ASR9K', 'CRS'}
    phases = {'Pre-Upgrade', 'Pre-Activate'}

    def run(self):
        """
        RP/0/RP0/CPU0:CRS-X-Deploy2#admin show redundancy location all
        Tue May 17 21:00:15.863 UTC
        Redundancy information for node 0/RP0/CPU0:
        ==========================================
        Node 0/RP0/CPU0 is in ACTIVE role
        Partner node (0/RP1/CPU0) is in STANDBY role
        Standby node in 0/RP1/CPU0 is ready
        Standby node in 0/RP1/CPU0 is NSR-ready
        """
        cmd = "show redundancy location all"
        if self.ctx.os_type != "eXR":
            cmd = "admin " + cmd
        output = self.ctx.send(cmd)
        lines = output.split("\n", 50)
        if len(lines) < 6:
            self.ctx.error("Show redundancy output is insufficient.")

        for ln, line in enumerate(lines[:6]):
            if "is in STANDBY role" in line:
                if "is ready" in lines[ln + 1]:
                    break
                else:
                    self.ctx.error("Standby is not ready. Upgrade can not proceed.")

        self.ctx.info("Node redundancy level sufficient")



"""This is a module providing FSM actions for Install Operations plugins."""


def a_error(plugin_ctx, ctx):
    """Display the error message."""
    message = ctx.ctrl.after.strip().splitlines()[-1]
    plugin_ctx.error(message)
    return False



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import re
import string

install_error_pattern = re.compile("Error:    (.*)$", re.MULTILINE)


def log_install_errors(ctx, output):
    """
    Print warning for Error:

    :param ctx:
    :param output:
    :return: nothing
    """
    errors = re.findall(install_error_pattern, output)
    for line in errors:
        ctx.warning(line)


def number_of_rsp(ctx):
    """
    Determine the number of RSP's in the chassis

    :param ctx:
    :return: the number of RSP's
    """
    platforms = ['ASR-902', 'ASR-920']
    count = 0
    valid_count = ['1', '2']

    if ctx._connection.platform in platforms:
        count = 1
        return count

    # show platform can take more than 1 minute after router reload. Issue No. 47
    output = ctx.send("show platform | count RSP", timeout=600)
    if output:
        m = re.search('Number.*= (\d+)', output)
        if m:
            count = m.group(1)
            if count not in valid_count:
                ctx.error("Invalid RSP count: {}".format(count))
            else:
                count = int(count)

    return count


def install_folder(ctx):
    """
    Determine the image folder
        'File: bootflash:/Image/packages.conf, on: RP0'
        'File: consolidated:packages.conf, on: RP0'

    :param   ctx
    :return: the image folder
    """

    folder = 'bootflash:/Image'

    output = ctx.send("show version running | include packages.conf")
    if output:
        m = re.search('File: (.*)/?packages.conf', output)
        if m:
            folder = m.group(1)
            folder = re.sub("/$", "", folder)
            if folder == 'consolidated:':
                folder = 'bootflash:/Image'

    return folder


def create_folder(ctx, folder):
    """
    Determine the image folder
        'File: bootflash:/Image/packages.conf, on: RP0'
        'File: consolidated:packages.conf, on: RP0'

    :param   ctx
    :param   folder to be created
    :return: True: Success, False: Failed
    """

    output = ctx.send('dir ' + folder)
    m = re.search('%Error opening', output)
    if m:
        cmd = 'mkdir ' + folder
        ctx.send(cmd, wait_for_string="Create directory filename")
        ctx.send('\r\n')
    else:
        return True

    output = ctx.send('dir ' + folder)
    m = re.search('%Error opening', output)
    if m:
        return False
    else:
        return True


def available_space(ctx, device):
    """
    Determine the available space on device such as bootflash or stby-bootflash:

    :param ctx:
    :param device: bootflash / stby-bootflash:
    :return: the available space
    """

    available = -1
    output = ctx.send('dir ' + device)
    m = re.search('(\d+) bytes free', output)
    if m:
        available = int(m.group(1))

    return available


def installed_package_name(ctx, pkg_conf):
    """
    :param: ctx
    :param: pkg_conf such as bootflash:/Image/packages.conf
    :return: the installed package name
    """
    output = ctx.send('dir ' + pkg_conf)
    if not output:
        ctx.error("dir {} failed".format(pkg_conf))
        return None

    m = re.search('No such file', output)
    if m:
        ctx.info('{} does not exist'.format(pkg_conf))
        return None

    cmd = "more " + pkg_conf + " | include PackageName"
    output = ctx.send(cmd)
    m = re.search('pkginfo: PackageName: (.*)$', output)
    if m:
        img_name = m.group(1)
        ctx.info("installed package name = {} ".format(img_name))
        return img_name
    else:
        ctx.info("PackageName is not found in {}".format(pkg_conf))
        return None


def installed_package_version(ctx):
    """
    :param: ctx
    :return: the installed package name
    """
    # cmd = "more " + pkg_conf + " | include Build:"
    # pkginfo: Build: 03.14.03.S.155-1.S3-std
    # output = ctx.send(cmd)
    # m = re.search('pkginfo: Build: (.*)$', output)

    cmd = 'show version | include Cisco IOS XE Software'
    output = ctx.send(cmd)
    # Cisco IOS XE Software, Version 03.13.03.S - Extended Support Release
    m = re.search('Version (.*) -', output)

    # Cisco IOS XE Software, Version 03.18.01.SP.156-2.SP1-ext
    if not m:
        m = re.search('Version (.*)$', output)

    if m:
        bld_version = m.group(1)
        ctx.info("installed package version = {}".format(bld_version))
        return bld_version
    else:
        ctx.warning("Build version is not found in show version: {}".format(output))
        return None


def installed_package_device(ctx):
    """
    :param: ctx
    :return: device_type with rsp version ie asr900rsp2
    """
    cmd = 'show version running | include File:'
    # File: bootflash:/Image/asr900rsp2-rpbase.03.13.03.S.154-3.S3-ext.pkg, on: RP0
    # File: bootflash:/Image/asr900rsp3-rpbase.V156_2_SP_SR680887239_5.pkg, on: RP0
    img_dev = None
    output = ctx.send(cmd)
    if output:
        lines = string.split(output, '\n')
        lines = [x for x in lines if x]
        for line in lines:
            m = re.search('File: .*(asr\w+)-\w+.\w+', line)
            if m:
                img_dev = m.group(1)
                break

    ctx.info("installed package device = {}".format(img_dev))
    return img_dev


def install_package_family(pkg):
    """
    :param: pkg ie asr900rsp2-universal.03.13.03.S.154-3.S3-ext.bin
    :return: device_type of the installed image ie asr900
    """
    img_dev = None
    m = re.search('(asr\d+)\w*', pkg)
    if m:
        img_dev = m.group(1)

    return img_dev


def install_add_remove(ctx, cmd):
    """
    Execute the copy command

    :param ctx
    :param cmd
    :return: nothing
    """
    message = "Waiting the operation to continue"
    ctx.info(message)
    ctx.post_status(message)

    ctx.send(cmd, wait_for_string="Destination filename")
    output = ctx.send("\r\n\r\n\r\n", timeout=3600)

    result = re.search("\d+ bytes copied in .* secs", output)

    if result:
        ctx.info("Command {} finished successfully".format(cmd))
        return
    else:
        log_install_errors(ctx, output)
        ctx.error("Command {} failed".format(cmd))


def check_pkg_conf(ctx, folder):
    """
    Remove the existing packages

    :param ctx
    :param folder: i.e. bootflash:/Image
    :return: True or False
    """
    pkg_conf = folder + '/packages.conf'
    output = ctx.send('more ' + pkg_conf + ' | include pkg$')

    if not output:
        return False

    lines = string.split(output, '\n')
    lines = [x for x in lines if x]
    for line in lines:
        if ctx._connection.os_version not in line:
            return False

    return True


def remove_exist_image(ctx, package):
    """
    Remove the existing packages

    :param ctx
    :param package
    :return: True or False
    """

    output = ctx.send('dir ' + package)
    m = re.search('No such file', output)

    if m:
        return True
    else:
        cmd = "del /force {}".format(package)
        ctx.send(cmd)
        ctx.info("Removing files : {}".format(package))

    output = ctx.send(cmd)
    m = re.search('No such file', output)
    if m:
        return True
    else:
        return False


def remove_exist_subpkgs(ctx, folder, pkg):
    """
    Remove residual packages from the earlier installations

    :param ctx
    :param folder: i.e. bootflash:/Image
    :return: True or False
    """

    pkg_conf = folder + '/packages.conf'
    # Skip if no packages.conf
    output = ctx.send('dir ' + pkg_conf)
    if not output:
        ctx.error("dir {} failed".format(pkg_conf))
        return

    m = re.search('No such file', output)
    if m:
        ctx.info('Booted from consolidated mode: '
                 '{} does not exist'.format(pkg_conf))
        return

    # Discover package name, version, and image device
    img_name = installed_package_name(ctx, pkg_conf)
    bld_version = installed_package_version(ctx)
    img_device = installed_package_device(ctx)

    if not bld_version or not img_device or not img_name:
        ctx.warning("Not able to determine package name, version, or image device.")
        ctx.warning("Packages left over from earlier installations will not be removed.")
        return

    # Remove all the bin files except the current install pkg

    if folder != 'bootflash:':
        package_name = folder + '/asr*.bin'
        remove_exist_image(ctx, package_name)
    else:
        package_name = folder + '*.bin'
        output = ctx.send('dir ' + package_name + ' | include bin')
        if not output:
            ctx.error("dir {} failed".format(package_name))
            return

        lines = string.split(output, '\n')
        lines = [x for x in lines if x]
        for line in lines:
            m = re.search('(asr.*\.bin)', line)
            if m:
                previous_pkg = m.group(0)
                if previous_pkg != pkg:
                    previous_package = folder + '/' + previous_pkg
                    remove_exist_image(ctx, previous_package)

    # Remove the packages.conf*- file
    package_name = folder + '/packages.conf*-'
    remove_exist_image(ctx, package_name)

    # Remove residual asr900*.conf
    package_name = folder + '/asr9*.conf'
    remove_exist_image(ctx, package_name)

    # Remove .pkg files
    cmd = 'dir ' + folder + '/*.pkg | include pkg'
    # Directory of bootflash:/Image/*.pkg
    # 15107  -rw-    41534024   Sep 8 2016 03:55:47 +00:00  asr900rsp2-espbase.03.14.03.S.155-1.S3-std.pkg
    output = ctx.send(cmd)

    if not output:
        return

    lines = string.split(output, '\n')
    lines = [x for x in lines if x]
    for line in lines:
        m = re.search('(asr9.*pkg)', line)
        if m:
            exfile = m.group(1)
            package = folder + '/' + exfile
            if bld_version not in package or img_device not in package:
                remove_exist_image(ctx, package)

    return


def check_issu_readiness(ctx, pkg, image_size):
    """
    Expand the consolidated file into the image folder

    :param: ctx
    :param: pkg
    :param: image_size
    :return: True or False
    """

    # check the current package mode
    cmd = 'show version | count packages.conf'
    output = ctx.send(cmd)
    if output:
        m = re.search('Number.*= (\d+)', output)
        if m:
            count = m.group(1)
            if count == '0':
                ctx.info("The current boot mode is consolidated package.")
                return False
        else:
            ctx.warning("Invalid show version output: {}".format(output))
            return False
    else:
        ctx.warning("Show version command error!")
        return False

    # check software compatibility
    cmd = 'show version | include System image file'
    output = ctx.send(cmd)
    if output:
        m = re.search('System image file is \"(.*)\"', output)
        if m:
            pkg_conf = m.group(1)
            img_name = installed_package_name(ctx, pkg_conf)
            if not img_name:
                ctx.warning("Installed package name {} is not found.".format(pkg_conf))
                return False
        else:
            ctx.warning("Show version command error!")
            return False
    else:
        ctx.warning("Show version command error!")
        return False

    m = re.search('asr\w+-(\w+)\.\w+', pkg)
    if m:
        pkg_name = m.group(1)
        if img_name != pkg_name:
            ctx.info("Incompatible packages: {} vs. {}".format(img_name, pkg_name))
            return False
    else:
        ctx.warning("Package name is not found in {}".format(pkg))
        return False

    # check image types between RSP's
    cmd = 'show version rp active running | include Package'
    output = ctx.send(cmd)
    cmd = 'show version rp standby running | include Package'
    stby_output = ctx.send(cmd)
    if output and stby_output:
        lines = string.split(output, '\n')
        lines = [x for x in lines if x]
        # Package: rpbase, version: 03.16.00.S.155-3.S-ext, status: active
        for line in lines:
            m = re.search('Package: (.*) status', line)
            if m:
                img_type = m.group(1)
                if img_type not in stby_output:
                    ctx.warning("Mismatched image types:")
                    ctx.warning("Active rp version: {}".format(output))
                    ctx.warning("Standby rp version: {}".format(stby_output))
                    return False
            else:
                ctx.warning("Invalid package version format: {}".format(line))
                return False
    else:
        ctx.warning("Show version command error!")
        return False

    # check the required disk space for ISSU
    # bootflash: requires additional 250 MB
    # stby-bootflash: requires additional 450 MB
    total_size = 250000000 + image_size
    flash_free = available_space(ctx, 'bootflash:')
    if flash_free < total_size:
        ctx.info("Total required / bootflash "
                 "available: {} / {} bytes".format(total_size, flash_free))
        ctx.info("Not enough space in bootflash: to perform ISSU. "
                 "Setting the Router to boot in sub-package mode.")
        return False

    total_size = 450000000 + image_size
    flash_free = available_space(ctx, 'stby-bootflash:')
    if flash_free < total_size:
        ctx.info("Total required / stby-bootflash "
                 "available: {} / {} bytes".format(total_size, flash_free))
        ctx.info("Not enough space in stby-bootflash: to perform ISSU. "
                 "Setting the Router to boot in sub-package mode.")
        return False
    else:
        ctx.info("There is enough space on bootflash and stby-bootflash to perform ISSU")

    # check show redundancy
    cmd = 'show redundancy | include Configured Redundancy Mode'
    output = ctx.send(cmd)
    if output:
        m = re.search('Configured Redundancy Mode = (.*)', output)
        if m:
            configed_mode = m.group(1)
            if configed_mode != 'sso':
                ctx.warning("Configured Redundancy Mode = {}".format(configed_mode))
                return False
        else:
            ctx.warning("Show redundancy command error!")
            return False
    else:
        ctx.warning("Show redundancy command error!")
        return False

    cmd = 'show redundancy | include Operating Redundancy Mode'
    output = ctx.send(cmd)
    if output:
        m = re.search('Operating Redundancy Mode = (.*)', output)
        if m:
            operating_mode = m.group(1)
            if operating_mode != 'sso':
                ctx.warning("Operating Redundancy Mode = {}".format(operating_mode))
                return False
        else:
            ctx.warning("Show redundancy command error!")
            return False
    else:
        ctx.warning("Show redundancy command error!")
        return False

    cmd = 'show redundancy | include Current Software state'
    output = ctx.send(cmd)
    if output:
        lines = string.split(output, '\n')
        lines = [x for x in lines if x]
        num_of_line = len(lines)
        if num_of_line != 2:
            ctx.warning("num_of_line = {}".format(num_of_line))
            ctx.warning("Current Software state = {}".format(output))
            return False

        m = re.search('Current Software state = (.*)', lines[0])
        if m:
            active_state = m.group(1)
            if 'ACTIVE' not in active_state:
                ctx.warning("show redundancy Active state check has failed")
                ctx.warning("active_state = {}".format(active_state))
                ctx.warning("Current Software state = {}".format(lines[0]))
                return False
        else:
            ctx.warning("Show redundancy command error!")
            return False

        m = re.search('Current Software state = (.*)', lines[1])
        if m:
            stby_state = m.group(1)
            if 'STANDBY HOT' not in stby_state:
                ctx.warning("show redundancy STANDBY HOT state check has failed")
                ctx.warning("stby_state = {}".format(stby_state))
                ctx.warning("Current Software state = {}".format(lines[1]))
                return False
        else:
            ctx.warning("Show redundancy command error!")
            return False

    else:
        ctx.warning("Show redundancy command error!")
        return False

    return True


def xe_show_platform(ctx):
    """
    Parse show platform output to extract the RP and SIP status
    :param: ctx
    :return: dictionary

    0         1         2         3         4         5         6
    012345678901234567890123456789012345678901234567890123456789012345678
    Slot      Type                State                 Insert time (ago)
    --------- ------------------- --------------------- -----------------
     0/0      12xGE-2x10GE-FIXED  ok                    15:09:04
    R1        A900-RSP2A-128      ok, active            14:09:30
    """
    platform_info = {}
    cmd = 'show platform'
    # show platform can take more than 1 minute after router reload. Issue No. 47
    output = ctx.send(cmd, timeout=600)
    if output:
        lines = string.split(output, '\n')
        lines = [x for x in lines if x]
        sip0 = False
        for line in lines:
            if not sip0:
                m = re.search('--------- ------------------- '
                              '--------------------- -----------------', line)
                if m:
                    sip0 = True
                continue

            m = re.search('Slot      CPLD Version        Firmware Version', line)
            if m:
                break

            Slot = line[:8].strip()
            Type = line[10:28].strip()
            State = line[30:50].strip()

            m1 = re.search('^0\/\d+', Slot)
            m2 = re.search('^R\d+', Slot)
            if m1 or m2:
                platform_info[Slot] = [Type, State]

    return platform_info



# =============================================================================
# plugin
#
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

from csmpe.plugins import CSMPlugin


class Plugin(CSMPlugin):
    """This plugin checks the configuration filesystem"""
    name = "Config Filesystem Check Plugin"
    platforms = {'ASR9K', 'CRS', 'NCS1K', 'NCS4K', 'NCS5K', 'NCS5500', 'NCS6K', 'IOSXRv-9K', 'IOSXRv-X64'}
    phases = {'Pre-Upgrade', "Pre-Activate", "Pre-Deactivate"}

    def run(self):
        """
        RP/0/RSP0/CPU0:R3#cfs check
        Tue May 17 09:56:43.720 UTC

        Creating any missing directories in Configuration File system...OK
        Initializing Configuration Version Manager...OK
        Syncing commit database with running configuration...OK
        """
        ok = 0
        message = []
        output = self.ctx.send("cfs check", timeout=300)
        lines = output.split("\n", 50)
        for line in lines:
            if line != "":
                message.append(line)
            if 'OK' in line:
                ok += 1

        for line in message:
            if ok < 3:
                self.ctx.warning(line)
            else:
                self.ctx.info(line)
        if ok < 3:
            self.ctx.error("The configuration filesystem has inconsistencies")
        else:
            self.ctx.info("Configuration filesystem is consistent")



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================


from package_lib import SoftwarePackage
from csmpe.plugins import CSMPlugin
from install import install_activate_deactivate
from install import send_admin_cmd
from install import check_ncs6k_release, check_ncs4k_release
from csmpe.core_plugins.csm_get_inventory.exr.plugin import get_package, get_inventory
from csmpe.core_plugins.csm_install_operations.utils import update_device_info_udi, get_cmd_for_install_activate_deactivate


class Plugin(CSMPlugin):
    """This plugin Activates packages on the device."""
    name = "Install Activate Plugin"
    platforms = {'ASR9K', 'NCS1K', 'NCS4K', 'NCS5K', 'NCS5500', 'NCS6K', 'IOSXRv-9K', 'IOSXRv-X64'}
    phases = {'Activate'}
    os = {'eXR'}

    def get_tobe_activated_pkg_list(self):
        """
        Produces a list of packaged to be activated
        """
        packages = self.ctx.software_packages

        pkgs = SoftwarePackage.from_package_list(packages)

        # RP/0/RP0/CPU0:Deploy#show install inactive
        # 5 inactive package(s) found:
        #     ncs6k-k9sec-5.2.5.47I
        #     ncs6k-mpls-5.2.5.47I
        #     ncs6k-5.2.5.47I.CSCuy47880-0.0.4.i
        #     ncs6k-mgbl-5.2.5.47I
        #     ncs6k-5.2.5.CSCuz65240-1.0.0

        admin_installed_inact = SoftwarePackage.from_show_cmd(send_admin_cmd(self.ctx, "show install inactive"))
        admin_installed_act = SoftwarePackage.from_show_cmd(send_admin_cmd(self.ctx, "show install active"))

        installed_inact = SoftwarePackage.from_show_cmd(self.ctx.send("show install inactive"))
        installed_act = SoftwarePackage.from_show_cmd(self.ctx.send("show install active"))

        installed_inact.update(admin_installed_inact)
        installed_act.update(admin_installed_act)

        # Packages to activate but not already active
        pkgs = pkgs - installed_act
        if pkgs:
            packages_to_activate = set()
            # After the packages are considered equal according to SoftwarePackage.__eq__(),
            # Use the package name in the inactive area.  It is possible that the package
            # name given for Activation may be an external filename like below.
            # ncs6k-5.2.5.CSCuz65240.smu to ncs6k-5.2.5.CSCuz65240-1.0.0
            for inactive_pkg in installed_inact:
                for pkg in pkgs:
                    if pkg == inactive_pkg:
                        packages_to_activate.add(inactive_pkg)

            if not packages_to_activate:
                to_activate = " ".join(map(str, pkgs))

                state_of_packages = "\nTo activate :{} \nInactive: {} \nActive: {}".format(
                    to_activate, installed_inact, installed_act
                )
                self.ctx.info(state_of_packages)
                self.ctx.error('To be activated packages not in inactive packages list.')
                return None
            else:
                if len(packages_to_activate) != len(packages):
                    self.ctx.info('Packages selected for activation: {}\n'.format(" ".join(map(str, packages))) +
                                  'Packages that are to be activated: {}'.format(" ".join(map(str,
                                                                                              packages_to_activate))))
                return " ".join(map(str, packages_to_activate))

    def run(self):
        """
        Performs install activate operation
        """
        check_ncs6k_release(self.ctx)
        check_ncs4k_release(self.ctx)

        cmd = get_cmd_for_install_activate_deactivate(self.ctx, self.get_tobe_activated_pkg_list,
                                                      cmd_with_package_names='install activate {}',
                                                      cmd_with_operation_id='install activate id {}')
        if not cmd:
            self.ctx.info("Nothing to be activated.")
            return True

        self.ctx.info("Activate package(s) pending")
        self.ctx.post_status("Activate Package(s) Pending")

        install_activate_deactivate(self.ctx, cmd)

        self.ctx.info("Activate package(s) done")
        self.ctx.info("Refreshing package and inventory information")
        self.ctx.post_status("Refreshing package and inventory information")
        # Refresh package and inventory information
        get_package(self.ctx)
        get_inventory(self.ctx)

        update_device_info_udi(self.ctx)
